#!/usr/bin/python
# -*- coding: utf-8 -*-
# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

#
#  _  __          _                               _        
# | |/ /   __ _  | |  _ __ ___     __ _    __ _  (_)   ___ 
# | ' /   / _` | | | | '_ ` _ \   / _` |  / _` | | |  / __|
# | . \  | (_| | | | | | | | | | | (_| | | (_| | | | | (__ 
# |_|\_\  \__, | |_| |_| |_| |_|  \__,_|  \__, | |_|  \___|
#            |_|                          |___/            
# binary:
#
# 01001011 01110001 01101100 01101101 01100001 01100111 01101001 01100011 
#
# morse:
#
# -.- --.- .-.. -- .- --. .. -.-. 
# 
# towpoint:
#
# |/(||._ _  _ (~|o _
# |\ ||| | |(_| _||(_
#
# italic
#
#  /__/  _  /  _   _   _  '  _ 
# /  )  (/ (  //) (/  (/ /  (  
#       /            _/     
#
#
# goofy:
#
# \   | )  /  /    | \   |    |        |    /  \      )  ____)  (_    _)  /  __) 
#  |  |/  /  (     |  |  |    |  |\/|  |   /    \    /  /  __     |  |   |  /    
#  |     (    \__  |  |  |    |  |  |  |  /  ()  \  (  (  (  \    |  |   | |     
#  |  |\  \      | |  |  |__  |  |  |  | |   __   |  \  \__)  )  _|  |_  |  \__  
# /   |_)  \_____| |_/      )_|  |__|  |_|  (__)  |___)      (__(      )__\    )_
#
# efiwater
#
#  _       _                  o     
#  )L7 __  )) _  _  ___  ___  _  __ 
# ((`\((_)(( ((`1( ((_( ((_( (( ((_ 
#       ))                _))      
# 
# rev:
#
# ========================================================
# =  ====  =========  ====================================
# =  ===  ==========  ====================================
# =  ==  ===========  ====================================
# =  =  ======    ==  ==  =  = ====   ====   ===  ===   ==
# =     =====  =  ==  ==        ==  =  ==  =  ======  =  =
# =  ==  ====  =  ==  ==  =  =  =====  ===    ==  ==  ====
# =  ===  ====    ==  ==  =  =  ===    =====  ==  ==  ====
# =  ====  =====  ==  ==  =  =  ==  =  ==  =  ==  ==  =  =
# =  ====  =====  ==  ==  =  =  ===    ===   ===  ===   ==
# ========================================================
#

                                                                                                                                          
                                                                                                                                          
# KKKKKKKKK    KKKKKKK                    lllllll                                                                 iiii                      
# K:::::::K    K:::::K                    l:::::l                                                                i::::i                     
# K:::::::K    K:::::K                    l:::::l                                                                 iiii                      
# K:::::::K   K::::::K                    l:::::l                                                                                           
# KK::::::K  K:::::KKK   qqqqqqqqq   qqqqq l::::l    mmmmmmm    mmmmmmm     aaaaaaaaaaaaa      ggggggggg   gggggiiiiiii     cccccccccccccccc
#   K:::::K K:::::K     q:::::::::qqq::::q l::::l  mm:::::::m  m:::::::mm   a::::::::::::a    g:::::::::ggg::::gi:::::i   cc:::::::::::::::c
#   K::::::K:::::K     q:::::::::::::::::q l::::l m::::::::::mm::::::::::m  aaaaaaaaa:::::a  g:::::::::::::::::g i::::i  c:::::::::::::::::c
#   K:::::::::::K     q::::::qqqqq::::::qq l::::l m::::::::::::::::::::::m           a::::a g::::::ggggg::::::gg i::::i c:::::::cccccc:::::c
#   K:::::::::::K     q:::::q     q:::::q  l::::l m:::::mmm::::::mmm:::::m    aaaaaaa:::::a g:::::g     g:::::g  i::::i c::::::c     ccccccc
#   K::::::K:::::K    q:::::q     q:::::q  l::::l m::::m   m::::m   m::::m  aa::::::::::::a g:::::g     g:::::g  i::::i c:::::c             
#   K:::::K K:::::K   q:::::q     q:::::q  l::::l m::::m   m::::m   m::::m a::::aaaa::::::a g:::::g     g:::::g  i::::i c:::::c             
# KK::::::K  K:::::KKKq::::::q    q:::::q  l::::l m::::m   m::::m   m::::ma::::a    a:::::a g::::::g    g:::::g  i::::i c::::::c     ccccccc
# K:::::::K   K::::::Kq:::::::qqqqq:::::q l::::::lm::::m   m::::m   m::::ma::::a    a:::::a g:::::::ggggg:::::g i::::::ic:::::::cccccc:::::c
# K:::::::K    K:::::K q::::::::::::::::q l::::::lm::::m   m::::m   m::::ma:::::aaaa::::::a  g::::::::::::::::g i::::::i c:::::::::::::::::c
# K:::::::K    K:::::K  qq::::::::::::::q l::::::lm::::m   m::::m   m::::m a::::::::::aa:::a  gg::::::::::::::g i::::::i  cc:::::::::::::::c
# KKKKKKKKK    KKKKKKK    qqqqqqqq::::::q llllllllmmmmmm   mmmmmm   mmmmmm  aaaaaaaaaa  aaaa    gggggggg::::::g iiiiiiii    cccccccccccccccc
#                                 q:::::q                                                               g:::::g                             
#                                 q:::::q                                                   gggggg      g:::::g                             
#                                q:::::::q                                                  g:::::gg   gg:::::g                             
#                                q:::::::q                                                   g::::::ggg:::::::g                             
#                                q:::::::q                                                    gg:::::::::::::g                              
#                                qqqqqqqqq                                                      ggg::::::ggg                                
#                                                                                                  gggggg                                   

"""A module that manage package version.
"""

import sys
import re
from functools import cmp_to_key
from typing import Any, Dict, Tuple, Iterable


from .constants import Constants
from .help import MarkdownString
from ._version import __version__
from .http_client import HttpClient


_IGNORE_POST_VERSION_PATTERN = r'[.]?(post|rev|r)[0-9]*$'
_IS_STABLE_VESRION_PATTERN = r'[v]?([0-9]+[!])?[0-9]+(\.[0-9]*)*([.]?(post|rev|r)[0-9]*)?$'

try:
    import pkg_resources

    def is_stable_version(version:str)->bool:
        parsed_version = pkg_resources.parse_version(version)
        return not parsed_version.is_prerelease  # is_prerelease returns true for .dev or .rc versions


    def compare_version(other:str, version:str, ignore_current_version_post:bool)->int:
        """ Compares current version to another version string.

        Parameters
        ----------
        other : str
            The other version to compare with, assume string "X.Y.Z" X,Y,Z integers
        version : str
            The current version to compare with, assume string "X.Y.Z" X,Y,Z integers
        ignore_current_version_post : bool
            If set the comparison should ignore current version post version information


        Returns
        -------
        int
            -1 if version higher than other
             0 if version equal to other
             1 if version lower than other
        """
        VERSION_BIGGER = -1
        VERSION_LOWER  =  1
        VERSION_EQUAL  =  0

        if ignore_current_version_post:
            version = re.sub(_IGNORE_POST_VERSION_PATTERN, '',version, re.IGNORECASE)
            other = re.sub(_IGNORE_POST_VERSION_PATTERN, '',other, re.IGNORECASE)
        if other == version:
            return VERSION_EQUAL

        other_parsed = pkg_resources.parse_version(other)
        version_parsed = pkg_resources.parse_version(version)
        if other_parsed == version_parsed:
            return VERSION_EQUAL

        if other_parsed < version_parsed:
            return VERSION_BIGGER
        else:
            return VERSION_LOWER

except:
    def is_stable_version(version:str)->bool:
        match = re.match(_IS_STABLE_VESRION_PATTERN, version, re.IGNORECASE)
        return match is not None


    def compare_version(other:str, version:str, ignore_current_version_post:bool=False)->int:
        """ Compares current version to another version string.

        Parameters
        ----------
        other : str
            The other version to compare with, assume string "X.Y.Z" X,Y,Z integers
        version : str
            The current version to compare with, assume string "X.Y.Z" X,Y,Z integers
        ignore_current_version_post : bool
            If set the comparison should ignore current version post versions


        Returns
        -------
        int
            -1 if version higher than other
             0 if version equal to other
             1 if version lower than other
        """

        VERSION_BIGGER = -1
        VERSION_LOWER  =  1
        VERSION_EQUAL  =  0

        try:
            if ignore_current_version_post:
                version = re.sub(_IGNORE_POST_VERSION_PATTERN, '',version, re.IGNORECASE)
                other = re.sub(_IGNORE_POST_VERSION_PATTERN, '',other, re.IGNORECASE)

            if other != version:
                other_list, other_pre_post_list = _normalize_version(other)
                version_list, version_pre_post_list = _normalize_version(version)
                max_len = max(len(version_list), len(other_list))
                other_list = [*other_list, *([0] * max(0, max_len - len(other_list))), *other_pre_post_list]
                version_list = [*version_list, *([0] * max(0, max_len - len(version_list))), *version_pre_post_list]

                for idx in range(0, max_len):
                    v_element = int(version_list[idx])
                    o_element = int(other_list[idx])
                    if v_element != o_element:
                        if v_element > o_element:
                            return VERSION_BIGGER
                        else:
                            return VERSION_LOWER
                                  
        except:
            pass
        return VERSION_EQUAL


    def _normalize_version(v:str)->Tuple[list,list]:
        PRE_POST_STR_TO_LEVEL = {
            "dev": -40,
            "a": -30,
            "alpha": -30,
            "b": -20,
            "beta": -20,
            "rc": -10,

            "post": 10,
            "r": 10,
            "rev": 10
        }
        v = v.strip().lower()
        v = v[1:] if v.startswith("v") else v
        
        version_list = v.split(".")

        first = version_list[0].strip()
        pair = first.split("!", 1)
        if len(pair) == 1:
            version_list = [0, *version_list]
        else:
            version_list = [*pair, *version_list[1:]]
        last = version_list[-1].strip()

        pre_post_list = [0,0]
        if not is_int(last):
            version_list = version_list[:-1]
            if last != "":
                match = re.match(r'([0-9]*)[\-_]?([a-z]*)[\-_]?([0-9]*)', last, re.IGNORECASE)
                if match:
                    groups = match.groups()
                    pre_post_level = PRE_POST_STR_TO_LEVEL.get(groups[1].lower())
                    if pre_post_level is not None:
                        if groups[0] != "":
                            version_list.append(groups[0])
                        pre_post_version = int(groups[2]) if groups[2] != "" else 0
                        pre_post_list = [pre_post_level, pre_post_version]
                else:
                    pre_post_list = [last, last]

        return version_list, pre_post_list


    def is_int(str_val:str)->bool:
        """ Checks whether a string can be converted to int.

        Parameters
        ----------
        str_val : str
            A string to be checked.

        Returns
        -------
        bool
            True if can be converted to int, otherwise False
        """

        return not (len(str_val) == 0 or any([c not in "0123456789" for c in str_val]))


    def to_int(str_val:str)->int:
        """ Converts string to int if possible.
        
        Parameters
        ----------
        str_val : str
            A string to be converted.

        Returns
        -------
        int or None
            Converted integer if success, otherwise None
        """

        return int(str_val) if is_int(str_val) else None


def pre_version_label(v:str)->str:
    PRE_STR_TO_LABEL = {
        "dev": "Development",
        "a": "Alpha",
        "alpha": "Alpha",
        "b": "Beta",
        "beta": "Beta",
        "rc": "Release Candidate",
    }
    version_list = v.split(".")
    last = version_list[-1].strip()
    if last != "":
        match = re.match(r'([0-9]*)[\-_]?([a-z]*)[\-_]?([0-9]*)', last, re.IGNORECASE)
        if match:
            groups = match.groups()
            return PRE_STR_TO_LABEL.get(groups[1])


def execute_version_command()->MarkdownString:
    """ execute the version command.
    command just return a string with the version that will be displayed to the user

    Returns
    -------
    str
        A string with the current version
    """

    return MarkdownString(f"{Constants.MAGIC_PACKAGE_NAME} version: {__version__}", title="version")


def get_pypi_latest_version(package_name:str)->Tuple[str,str]:
    """ Retreives latest package version string for PyPI.

    Parameters
    ----------
    package_name : str
        Name of package as define in PyPI.

    Returns
    -------
    str, str or None, None
        The latest version string and latest stable string of package in PyPI, or None, None if fails to retrieve

    Raises
    ------
    RequestsException
        If request to PyPI fails.
    """

    #
    # reterive package data
    #

    json_response = _retreive_package_from_pypi(package_name)
    all_releases = json_response.get("releases").keys()
    all_releases_desc_sorted = sorted(all_releases, key=cmp_to_key(lambda v,o: compare_version(v, o, False)), reverse=True)

    latest_stable_release = latest_release = json_response.get("info").get("version") if json_response and json_response.get("info") else None

    if len(all_releases_desc_sorted) > 0:
        latest_release = all_releases_desc_sorted[0]
        latest_stable_release = _get_latest_stable_version(all_releases_desc_sorted) or latest_stable_release or latest_release

    return latest_release, latest_stable_release


def _get_latest_stable_version(all_releases_desc_sorted:Iterable)->str:
    for release in all_releases_desc_sorted:
        if is_stable_version(release):
            return release


def _retreive_package_from_pypi(package_name:str)->Dict[str,Any]:
    """ Retreives package data from PyPI.

    Parameters
    ----------
    package_name : str
        Name of package as define in PyPI.

    Returns
    -------
    dict or None
        The package PyPI data, or None if fails to retrieve

    Raises
    ------
    RequestsException
        If request to PyPI fails.
    """

    api_url = f"https://pypi.org/pypi/{package_name}/json"
    http_client = HttpClient()
    response = http_client.get(api_url)

    #
    # handle response
    #

    response.raise_for_status()
    json_response = response.json()
    return json_response


def validate_required_python_version_running(minimal_required_version:str)->None:
    """ Validate whether the running python version meets minimal required python version 
    
    Parameters
    ----------
    minimal_required_version : str
        Minimal required python version, in the following format: major.minor.micro

    Returns
    -------
    None

    Exceptions
    ----------
    Raise RunTime exception, if sys.version_info does not support attributes: major, minor, micro (old python versions)
    Raise RunTime exception, if running python version is lower than required python version 

    """

    try:
        parts = minimal_required_version.split(".")
        min_py_version = 1000000 * int(parts[0]) + 1000 * (int(parts[1]) if len(parts) > 1 else 0) + (int(parts[2]) if len(parts) > 2 else 0)
        running_py_version = 1000000 * sys.version_info.major + 1000 * sys.version_info.minor + sys.version_info.micro
        if running_py_version < min_py_version:
            raise RuntimeError("")
    except:
        raise RuntimeError(f"Kqlmagic requires python >= {Constants.MINIMAL_PYTHON_VERSION_REQUIRED}, you use python {sys.version}")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Tuple, Dict, List, Any
import itertools
import configparser as CP
from datetime import timedelta, datetime

import dateutil.parser
from traitlets import TraitType
from traitlets.config.configurable import Configurable


from ._debug_utils import debug_print 
from .dependencies import Dependencies
from .constants import Constants, Schema
from .my_utils import split_lex, adjust_path, is_env_var, get_env_var, is_collection, strip_if_quoted
from .engine import Engine

from .commands_table import COMMANDS_TABLE


class Parser(object):

    default_options:Dict[str,Any] = {}
    traits_dict:Dict[str,TraitType] = {}

    @classmethod
    def initialize(cls, config:Configurable):
        cls.traits_dict = config.traits()
        config.observe(Parser.observe_config_changes)
        cls.init_default_options(config)

    @staticmethod
    def observe_config_changes(change:Dict[str,str]): 
        if change.get('type') == 'change':
            name = change.get('name')
            obj = Parser._OPTIONS_TABLE.get(name.lower().replace("-", "").replace("_", ""))
            if "init" not in obj:
                Parser.default_options[change.get('name')] = change.get('new')


    @classmethod
    def init_default_options(cls, config:Configurable):
        for obj in cls._OPTIONS_TABLE.values():
            if "abbreviation" not in obj:
                name = obj.get("flag")
                if "init" in obj:
                    cls.default_options[name] = obj.get("init")
                elif name in cls.traits_dict:
                    cls.default_options[name] = getattr(config, name)
                elif "abbreviation" not in obj:
                    raise Exception(f"missing init in _OPTIONS_TABLE for option: {name}")


    @classmethod
    def parse(cls, _line:str, _cell:str, config:Configurable, engines:List[Engine], user_ns:Dict[str,Any])->List[Dict[str,Any]]:
        is_cell = _cell is not None
        cell = f"{_line}\n{_cell or ''}"
        cell = cell.strip()
        code = cell
        #
        # split string to queries
        #
        suppress_all_results = False
        # tuple: 
        sections:List[Dict[str,str]] = []
        if is_cell:
            magic_section_name = Constants.MAGIC_NAME
            section_lines:List[str] = []
            previous_line = " "  # should be init to space for the below to work
            # note: splitlines don't remove the \n suffix, each line endswith \n
            for line in code.splitlines(True):
                lstripped_line = line.lstrip()

                if (lstripped_line.startswith(Constants.IPYKERNEL_CELL_MAGIC_PREFIX)
                        and (previous_line.isspace() or (len(sections) > 0 and sections[-1].get("type") == "line_magic"))):
                    if len(section_lines) > 0:
                        sections.append({"type": "cell_magic", "name": magic_section_name, "body": "".join(section_lines)})

                    magic_word = lstripped_line.split(None, 1)[0]
                    magic_section_name = magic_word[len(Constants.IPYKERNEL_CELL_MAGIC_PREFIX):]
                    lstripped_line = lstripped_line[len(magic_word):].lstrip()
                    if magic_section_name == Constants.MAGIC_NAME:
                        section_lines = [] if lstripped_line.isspace() else [lstripped_line]
                    else:
                        section_lines = [lstripped_line]

                elif magic_section_name != Constants.MAGIC_NAME:
                    section_lines.append(line)

                elif (lstripped_line.startswith(Constants.IPYKERNEL_LINE_MAGIC_PREFIX)
                      and (previous_line.isspace() or (len(sections) > 0 and sections[-1].get("type") == "line_magic"))):
                    magic_word = lstripped_line.split(None, 1)[0]
                    magic_name = magic_word[len(Constants.IPYKERNEL_LINE_MAGIC_PREFIX):]
                    lstripped_line = lstripped_line[len(magic_word):].lstrip()
                    if magic_name == Constants.MAGIC_NAME:
                        if not lstripped_line.isspace():
                            sections.append({"type": "line_magic", "name": magic_name, "body": lstripped_line})
                    else:
                        sections.append({"type": "line_magic", "name": magic_name, "body": lstripped_line})

                elif line.isspace():
                    if len(section_lines) > 0:
                        not_commented_lines = [1 for seg_line in section_lines if not seg_line.lstrip().startswith("//")]
                        if (len(not_commented_lines) > 0):
                            sections.append({"type": "cell_magic", "name": magic_section_name, "body": "".join(section_lines)})
                        section_lines = []

                else:
                    section_lines.append(line)

                previous_line = line

            if len(section_lines) > 0:
                if magic_section_name == Constants.MAGIC_NAME:
                    not_commented_lines = [1 for seg_line in section_lines if not seg_line.lstrip().startswith("//")]
                    if (len(not_commented_lines) > 0):
                        sections.append({"type": "cell_magic", "name": magic_section_name, "body": "".join(section_lines)})
                else:
                    sections.append({"type": "cell_magic", "name": magic_section_name, "body": "".join(section_lines)})

            if len(sections) > 0:
                last_query = sections[-1].get("body").strip()
                if last_query == ";":
                    suppress_all_results = True
                    sections = sections[:-1]

            if len(sections) == 0:
                sections.append({"type": "cell_magic", "name": Constants.MAGIC_NAME, "body": ""})

        else:
            sections.append({"type": "line_magic", "name": Constants.MAGIC_NAME, "body": code.strip()})

        #
        # parse code to kql and options
        #
        parsed_sections = []
        last_connection_string = ""
        for section in sections:
            parsed_section = cls._parse_one_section(section, is_cell, config, engines, user_ns)
            connection_string = parsed_section.get("connection_string")
            if connection_string:
                last_connection_string = connection_string
            elif len(parsed_section.get("command")) == 0:
                parsed_section["connection_string"] = last_connection_string
            if suppress_all_results:
                parsed_section.get("options")["suppress_results"] = True
            
            parsed_sections.append(parsed_section)

        if len(parsed_sections) > 0:
            parsed_sections[-1]["last_query"] = True

        return parsed_sections


    @classmethod
    def _parse_one_section(cls, section:Dict[str,str], is_cell:bool, config:Configurable, engines:List[Engine], user_ns:Dict[str,Any])->Dict[str,Any]:
        """Separate input into (connection info, KQL statements, options)"""

        cell, command = cls._parse_kql_command(section, user_ns)
        command_name = command.get("command")
        if command_name is not None and command_name != "submit":
            cell_rest, options = cls._parse_kql_options(cell.strip(), is_cell, config, user_ns)
            cell_rest = cls._update_kql_command_params(command, cell_rest, user_ns)
            cls._validate_kql_command_params(command)
            if cell_rest: 
                raise ValueError(f"command --{command_name} has too many parameters")
            parsed_query = {"connection_string": "", "query": "", "options": options, "command": command}
            return parsed_query

        # split to max 2 parts. First part, parts[0], is the first string.
        # parts = [part.strip() for part in cell.split(None, 1)]
        parts = split_lex(cell)
        
        # print(parts)
        if not parts:
            kql, options = cls._parse_kql_options("", is_cell, config, user_ns)
            parsed_query = {"connection_string": "", "query": kql, "options": options, "command": {}}
            return parsed_query

        #
        # replace substring of the form $name or ${name}, in windows also %name% if found in env variabes
        #

        connection_string = None

        conn_str = parts[0].strip()
        if not conn_str.startswith(('-', '+')):
            _was_quoted, conn_str = strip_if_quoted(conn_str)

            #
            # connection taken from a section in  dsn file (file name have to be define in config.dsn_filename or specified as a parameter)
            #
            if is_collection(conn_str, "["):
                section = conn_str[1:-1].strip()

                # parse to get flag, for the case that the file nema is specified in the options
                code = cell[len(parts[0]):]
                kql, options = cls._parse_kql_options(code, is_cell, config, user_ns)

                parser = CP.ConfigParser()
                dsn_filename = adjust_path(options.get("dsn_filename", config.dsn_filename))
                parser.read(dsn_filename)
                cfg_dict = dict(parser.items(section))

                cfg_dict_lower = {k.lower().replace("_", "").replace("-", ""): v for (k, v) in cfg_dict.items()}
                for e in engines:
                    if e.get_mandatory_key() in cfg_dict_lower.keys():
                        all_keys = set(itertools.chain(*e.get_valid_keys_combinations()))
                        connection_kv = [f"{k}='{v}'" for k, v in cfg_dict_lower.items() if v and k in all_keys]
                        connection_string = f"{e.get_uri_schema_name()}://{';'.join(connection_kv)}"
                        break

            #
            # connection specified starting with one of the supported prefixes
            #
            elif "://" in conn_str:
                sub_parts = conn_str.strip().split("://", 1)
                if (len(sub_parts) == 2 and sub_parts[0].lower().replace("_", "").replace("-", "") in list(itertools.chain(*[e._ALT_URI_SCHEMA_NAMES for e in engines]))):
                    connection_string = conn_str
            #
            # connection specified as database@cluster
            #
            elif "@" in conn_str and "|" not in conn_str and "'" not in conn_str and '"' not in conn_str:
                connection_string = conn_str
        #
        # connection not specified, override default
        #
        if connection_string is None:
            connection_string = ""
            code = cell
        else:
            code = cell[len(parts[0]):]


        #
        # parse code to kql and options
        #

        kql, options = cls._parse_kql_options(code.strip(), is_cell, config, user_ns)
        kql = options.pop("query", None) or kql
        connection_string = options.pop("conn", None) or connection_string

        parsed_query = {"connection_string": connection_string.strip(), "query": kql, "options": options, "command": {}}

        return parsed_query


    @classmethod
    def parse_old(cls, line:str, cell:str, config:Configurable, engines:List[Engine], user_ns:Dict[str,Any])->List[Dict[str,Any]]:
        """Separate input into (connection info, KQL statements, options)"""

        is_cell = cell is not None
        cell = f"{line}\n{cell or ''}"
        cell = cell.strip()
        parsed_queries = []
        cell, command = cls._parse_kql_command(cell, user_ns)
        command_name = command.get("command")
        if command_name is not None and command_name != "submit":
            cell_rest, options = cls._parse_kql_options(cell.strip(), is_cell, config, user_ns)
            cell_rest = cls._update_kql_command_params(command, cell_rest, user_ns)
            cls._validate_kql_command_params(command)
            if cell_rest: 
                raise ValueError(f"command --{command_name} has too many parameters")
            parsed_queries.append({"connection_string": "", "query": "", "options": options, "command": command})
            return parsed_queries

        # split to max 2 parts. First part, parts[0], is the first string.
        # parts = [part.strip() for part in cell.split(None, 1)]
        parts = split_lex(cell)
        
        # print(parts)
        if not parts:
            kql, options = cls._parse_kql_options("", is_cell, config, user_ns)
            parsed_queries.append({"connection_string": "", "query": kql, "options": options, "command": {}})
            return parsed_queries

        #
        # replace substring of the form $name or ${name}, in windows also %name% if found in env variabes
        #

        connection_string = None

        conn_str = parts[0].strip()
        if not conn_str.startswith(('-', '+')):
            _was_quoted, conn_str = strip_if_quoted(conn_str)

            #
            # connection taken from a section in  dsn file (file name have to be define in config.dsn_filename or specified as a parameter)
            #
            if is_collection(conn_str, "["):
                section = conn_str[1:-1].strip()

                # parse to get flag, for the case that the file nema is specified in the options
                code = cell[len(parts[0]):]
                kql, options = cls._parse_kql_options(code, is_cell, config, user_ns)

                parser = CP.ConfigParser()
                dsn_filename = adjust_path(options.get("dsn_filename", config.dsn_filename))
                parser.read(dsn_filename)
                cfg_dict = dict(parser.items(section))

                cfg_dict_lower = {k.lower().replace("_", "").replace("-", ""): v for (k, v) in cfg_dict.items()}
                for e in engines:
                    if e.get_mandatory_key() in cfg_dict_lower.keys():
                        all_keys = set(itertools.chain(*e.get_valid_keys_combinations()))
                        connection_kv = [f"{k}='{v}'" for k, v in cfg_dict_lower.items() if v and k in all_keys]
                        connection_string = f"{e.get_uri_schema_name()}://{';'.join(connection_kv)}"
                        break

            #
            # connection specified starting with one of the supported prefixes
            #
            elif "://" in conn_str:
                sub_parts = conn_str.strip().split("://", 1)
                if (len(sub_parts) == 2 and sub_parts[0].lower().replace("_", "").replace("-", "") in list(itertools.chain(*[e._ALT_URI_SCHEMA_NAMES for e in engines]))):
                    connection_string = conn_str
            #
            # connection specified as database@cluster
            #
            elif "@" in conn_str and "|" not in conn_str and "'" not in conn_str and '"' not in conn_str:
                connection_string = conn_str
        #
        # connection not specified, override default
        #
        if connection_string is None:
            connection_string = ""
            code = cell
        else:
            code = cell[len(parts[0]):]

        #
        # split string to queries
        #
        suppress_all_results = False
        queries:List[str] = []
        if is_cell:
            queryLines:List[str] = []
            last_line:str = None
            for last_line in code.splitlines(True):
                # note: splitlines don't remove the \n suffix, each line endswith \n 
                if last_line.isspace():
                    if len(queryLines) > 0:
                        queries.append("".join(queryLines))
                        queryLines = []
                else:
                    queryLines.append(last_line)

            if len(queryLines) > 0:
                queries.append("".join(queryLines))

            if len(queries) > 0:
                last_query = queries[-1].strip()
                if last_query == ";":
                    suppress_all_results = True
                    queries = queries[:-1]

            if len(queries) == 0:
                queries.append("")
        else:
            queries.append(code.strip())

        #
        # parse code to kql and options
        #
        for query in queries:
            kql, options = cls._parse_kql_options(query.strip(), is_cell, config, user_ns)
            kql = options.pop("query", None) or kql
            connection_string = options.pop("conn", None) or connection_string
            if suppress_all_results:
                options["suppress_results"] = True
            parsed_queries.append({"connection_string": connection_string.strip(), "query": kql, "options": options, "command": {}})

        return parsed_queries

    @classmethod
    def _parse_kql_command(cls, section:Dict[str,str], user_ns:Dict[str,Any])->Tuple[str,Dict[str,Any]]:
        code = section.get("body")
        if section.get("name") != Constants.MAGIC_NAME:
            name = section.get("name").replace("_", "").replace("-", "")
            if name in ["py", "pyro", "pyrw"]:
                obj = COMMANDS_TABLE.get(name)
                return ("", {"command": obj.get("flag"), "obj": obj, "params": [section.get("body"), name]})
            else:
                # line/cell magic
                name = section.get("type")
                obj = COMMANDS_TABLE.get(name.replace("_", ""))
                command_name = obj.get("flag")

                params = []
                body = section.get("body")
                if command_name == "cell_magic":
                    first_line = body.split("\n",1)[0]
                    params.append(first_line.strip())
                    body = body[len(first_line) + 1:]

                params.append(body)
                params.append(section.get("name"))

                return ("", {"command": command_name, "obj": obj, "params": params})

        # kql section
        lookup_key = None
        trimmed_code = code
        skip_words_count = 0
        obj = None
        params_type = None
        command_name = None
        params = []
        words = code.split()
        more_words_count = len(words)
        for word in words:
            more_words_count -= 1

            if params_type == "not_quoted_str" and not word.startswith("-"):
                # break
                pass

            if skip_words_count == 0:
                _comment, skip_words_count = cls._parse_comment(word, trimmed_code)

            if skip_words_count > 0:
                skip_words_count -= 1
                trimmed_code = trimmed_code[trimmed_code.find(word) + len(word):]
                continue

            # command 
            elif command_name is None:
                if not word.strip().startswith("--"):
                    break
                word = word[2:]
                if word.startswith("-"):
                    raise ValueError(f"unknown command --{word}, commands' prefix should be a double hyphen-minus, not a triple hyphen-minus")
                lookup_key = word.lower().replace("_", "").replace("-", "")
                obj = COMMANDS_TABLE.get(lookup_key)
                if obj is None:
                    raise ValueError(f"unknown command --{word}")
                command_name = obj.get("flag")
                trimmed_code = trimmed_code[trimmed_code.find(word) + len(word):]

                params_type = obj.get("type")

                if params_type is None:
                    break
            # option
            elif word.startswith("-"):
                break

            # command's parameters
            else:
                command = {"command": command_name, "obj": obj, "params": params}
                EXIT_ON_OPTION = True
                trimmed_code, params = cls._parse_kql_command_params(command, trimmed_code, EXIT_ON_OPTION, user_ns)
                break

        if command_name is None:
            return (code.strip(), {})

        if command_name == "python":
            params = params or [""]
            params.append(lookup_key) # type of python "py", "pyro", "pyrw"

        elif command_name in ["line_magic", "cell_magic"]:
            body = params[0] if params else ""
            magic_word = body.split(None, 1)[0]
            body = body.lstrip()[len(magic_word):]

            params = []
            if command_name == "cell_magic":
                first_line = body.split("\n", 1)[0]
                params.append(first_line.strip())
                body = body[len(first_line) + 1:]
            else:
                body = body.lstrip()
            params.append(body)

            ipykernel_prefix_len = 0
            if command_name == "cell_magic" and magic_word.startswith(Constants.IPYKERNEL_CELL_MAGIC_PREFIX):
                ipykernel_prefix_len = len(Constants.IPYKERNEL_CELL_MAGIC_PREFIX)
            elif command_name == "line_magic" and magic_word.startswith(Constants.IPYKERNEL_LINE_MAGIC_PREFIX):
                ipykernel_prefix_len = len(Constants.IPYKERNEL_LINE_MAGIC_PREFIX)
            magic_name = magic_word[ipykernel_prefix_len:]
            params.append(magic_name) # the magic name

            if magic_name in ["py", "pyro", "pyrw"]:
                obj = COMMANDS_TABLE.get(magic_name)
                body = params[0] if command_name == "line_magic" else f"{params[0]}\n{params[1]}"
                return ("", {"command": obj.get("flag"), "obj": obj, "params": [body, magic_name]})

        return (trimmed_code.strip(), {"command": command_name, "obj": obj, "params": params})


    @classmethod
    def _parse_kql_command_params(cls, command:dict, code:str, exit_on_options:bool, user_ns:Dict[str,Any])->Tuple[str,List[str]]:
        trimmed_code = code.strip()
        params = command.get("params")
        obj = command.get("obj")
        params_type = obj.get("type")
        if params_type == "not_quoted_str":
            if exit_on_options and trimmed_code.startswith("-"):
                # DO NOTHING, EXIT
                pass
            else:
                params.append(trimmed_code)
                trimmed_code = ""
        elif params_type is not None:
            skip_words_count = 0
            command_name = command.get("command")
            words = code.split()
            for word in words:
                if skip_words_count == 0:
                    _comment, skip_words_count = cls._parse_comment(word, trimmed_code)

                if skip_words_count > 0:
                    skip_words_count -= 1
                    trimmed_code = trimmed_code[trimmed_code.find(word) + len(word):]
                    continue

                # option
                if exit_on_options and word.startswith("-"):
                    break

                # command's parameters

                if params_type == "str" and word[0] in ["'",'"']:
                    quoted_string, skip_words_count = cls.parse_quote(trimmed_code)
                    param = quoted_string
                    skip_words_count -= 1
                else:
                    param = word
                params.append(cls._parse_value("command", obj, command_name, param, user_ns))
                trimmed_code = trimmed_code[trimmed_code.find(word) + len(word):]

        return trimmed_code.strip(), params


    @classmethod
    def _update_kql_command_params(cls, command:dict, cell_rest:str, user_ns:Dict[str,Any])->str:
        params = command.get("params")
        if len(params) == 0:
            obj = command.get("obj")
            if obj.get("type") == "not_quoted_str":
                params.append(cell_rest)
                cell_rest = ""

            elif len(cell_rest) > 0 and obj.get("type") is not None:
                DONT_EXIT_ON_OPTION = False
                cell_rest, params = cls._parse_kql_command_params(command, cell_rest, DONT_EXIT_ON_OPTION, user_ns)
                cell_rest = ""
            elif "default" in obj:
                params.append(obj.get("default"))
            command["params"] = params

        return cell_rest


    @classmethod
    def _validate_kql_command_params(cls, command:dict):
        params = command.get("params")
        command_name = command.get("command")
        obj = command.get("obj")
        _type = obj.get("type")

        max_params = obj.get("max_params") or (1 if _type is not None else 0)

        if len(params) > max_params:
            raise ValueError(f"command --{command_name} has too many parameters")

        min_params = obj.get("min_params") or (1 if _type is not None else 0)

        if len(params) < min_params:
            raise ValueError(f"command --{command_name} is missing parameter")


    @classmethod
    def validate_query_properties(cls, schema:str, properties:Dict[str,Any])->None:
        if type(properties) == dict:
            usupported_properties = []
            for p in properties:
                prop = cls._QUERY_PROPERTIES_TABLE[p]
                prop_schema_list = prop.get("schema")
                if type(prop_schema_list) == list and schema not in prop_schema_list and len(prop_schema_list) > 0 and schema is not None:
                    usupported_properties.append(p)
                if len(usupported_properties) > 0:
                    raise ValueError(f"query properties {usupported_properties} are not supported by current connection")


    _QUERY_PROPERTIES_TABLE = {
        # NOT DOCUMENTED - (OptionBlockSplittingEnabled): Enables splitting of sequence blocks after aggregation operator. [Boolean]
        "block_splitting_enabled": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # NOT DOCUMENTED - (OptionDatabasePattern): Database pattern overrides database name and picks the 1st database that matches the pattern. 
        # '*' means any database that user has access to. [String]
        "database_pattern": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},

        # NOT DOCUMENTED - If set, don't fuse projection into ExternalData operator. [bool]
        "debug_query_externaldata_projection_fusion_disabled": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # NOT DOCUMENTED - The percentage of threads to fanout execution to for external data nodes. [int]
        "debug_query_fanout_threads_percent_external_data":  {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},
        # (OptionDeferPartialQueryFailures): If true, disables reporting partial query failures as part of the result set. [Boolean]
        "deferpartialqueryfailures": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionMaterializedViewShuffleQuery): A hint to use shuffle strategy for materialized views that are referenced in the query. The property is an array of materialized views names and the shuffle keys to use. examples: 'dynamic([ { "Name": "V1", "Keys" : [ "K1", "K2" ] } ])' (shuffle view V1 by K1, K2) or 'dynamic([ { "Name": "V1" } ])' (shuffle view V1 by all keys) [dynamic]
        "materialized_view_shuffle": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "dict"},

        # (OptionMaxMemoryConsumptionPerQueryPerNode): Overrides the default maximum amount of memory a whole query may allocate per node. [UInt64]
        "max_memory_consumption_per_query_per_node": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # (OptionMaxMemoryConsumptionPerIterator): Overrides the default maximum amount of memory a query operator may allocate. [UInt64]
        "maxmemoryconsumptionperiterator": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # (OptionMaxOutputColumns): Overrides the default maximum number of columns a query is allowed to produce. [Long]
        "maxoutputcolumns": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # (OptionNoRequestTimeout): Enables setting the request timeout to its maximum value. [Boolean]
        "norequesttimeout": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionNoTruncation): Enables suppressing truncation of the query results returned to the caller. [Boolean]
        "notruncation": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionPushSelectionThroughAggregation): If true, push simple selection through aggregation [Boolean]
        "push_selection_through_aggregation": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # NOT DOCUMENTED - (OptionAdminSuperSlackerMode): If true, delegate execution of the query to another node [Boolean]
        "query_admin_super_slacker_mode": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (QueryBinAutoAt): When evaluating the bin_auto() function, the start value to use. [LiteralExpression]
        "query_bin_auto_at": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},

        # (QueryBinAutoSize): When evaluating the bin_auto() function, the bin size value to use. [LiteralExpression]
        "query_bin_auto_size": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},

        # (OptionQueryCursorAfterDefault): The default parameter value of the cursor_after() function when called without parameters. [string]
        "query_cursor_after_default": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},

        # NOT DOCUMENTED - (OptionQueryCursorAllowReferencingStreamingIngestionTables): Enable usage of cursor functions over databases which have streaming ingestion enabled. [boolean]
        "query_cursor_allow_referencing_streaming_ingestion_tables": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionQueryCursorBeforeOrAtDefault): The default parameter value of the cursor_before_or_at() function when called without parameters. [string]
        "query_cursor_before_or_at_default": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},

        # (OptionQueryCursorCurrent): Overrides the cursor value returned by the cursor_current() or current_cursor() functions. [string]
        "query_cursor_current": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},

        # (OptionQueryCursorDisabled): Disables usage of cursor functions in the context of the query. [boolean]
        "query_cursor_disabled ": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionQueryCursorScopedTables): List of table names that should be scoped to cursor_after_default .. 
        # cursor_before_or_at_default (upper bound is optional). [dynamic]
        "query_cursor_scoped_tables": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "dict"},

        # (OptionQueryDataScope): Controls the query's datascope -- whether the query applies to all data or just part of it. ['default', 'all', or 'hotcache']
        "query_datascope": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "enum", "values": ['default', 'all', 'hotcache']},

        # (OptionQueryDateTimeScopeColumn): Controls the column name for the query's datetime scope (query_datetimescope_to / query_datetimescope_from). [String]
        "query_datetimescope_column": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},

        # (OptionQueryDateTimeScopeFrom): Controls the query's datetime scope (earliest) 
        # used as auto-applied filter on query_datetimescope_column only (if defined). [DateTime]
        "query_datetimescope_from": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},

        # (OptionQueryDateTimeScopeTo): Controls the query's datetime scope (latest)
        # used as auto-applied filter on query_datetimescope_column only (if defined). [DateTime]
        "query_datetimescope_to": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},

        # (OptionQueryDistributionNodesSpanSize): If set, controls the way sub-query merge behaves: 
        # the executing node will introduce an additional level in the query hierarchy for each sub-group of nodes; the size of the sub-group is set by this option. [Int]
        "query_distribution_nodes_span": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "int"},

        # (OptionQueryFanoutNodesPercent): The percentage of nodes to fanour execution to. [Int]
        "query_fanout_nodes_percent": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # (OptionQueryFanoutThreadsPercent): The percentage of threads to fanout execution to. [Int]
        "query_fanout_threads_percent": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # (OptionQueryForceRowLevelSecurity): If specified, forces Row Level Security rules, even if row_level_security policy is disabled [Boolean]
        "query_force_row_level_security": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionQueryLanguage): Controls how the query text is to be interpreted. ['csl','kql' or 'sql']
        "query_language": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "enum", "values": ['csl', 'kql', 'sql']},

        # NOT DOCUMENTED - (RemoteMaterializeOperatorInCrossCluster): Enables remoting materialize operator in cross cluster query.
        "query_materialize_remote_subquery": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionMaxEntitiesToUnion): Overrides the default maximum number of columns a query is allowed to produce. [Long]
        "query_max_entities_in_union": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # (OptionQueryNow): Overrides the datetime value returned by the now(0s) function. [DateTime]
        # note: cannot be relative to now()
        "query_now": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},

        # (OptionDebugPython): If set, generate python debug query for the enumerated python node (default first). [Boolean or Int]
        "query_python_debug": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # (OptionQueryResultsApplyGetSchema): If set, retrieves the schema of each tabular data in the results of the query instead of the data itself. [Boolean]
        "query_results_apply_getschema": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionQueryResultsCacheMaxAge): If positive, controls the maximum age of the cached query results that Kusto is allowed to return [TimeSpan]
        "query_results_cache_max_age": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},
        
        # NOT DOCUMENTED - (CostBasedOptimizerBroadcastJoinBuildMax): Max Rows count for build in broadcast join.
        "query_optimization_broadcast_build_maxSize": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # NOT DOCUMENTED - (CostBasedOptimizerBroadcastJoinProbeMin): Min Rows count for probe in broadcast join.
        "query_optimization_broadcast_probe_minSize": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # NOT DOCUMENTED - (CostBasedOptimizer): Enables automatic optimizations.
        "query_optimization_costbased_enabled": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # NOT DOCUMENTED - (OptionOptimizeInOperator): Optimizes in operands serialization.
        "query_optimization_in_operator": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # NOT DOCUMENTED - (CostBasedOptimizerShufflingCardinalityThreshold): Shuffling Cardinality Threshold.
        "query_optimization_shuffling_cardinality": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # NOT DOCUMENTED - (OptionQueryRemoteEntitiesDisabled): If set, queries cannot access remote databases / clusters. [Boolean]
        "query_remote_entities_disabled": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # NOT DOCUMENTED - (RemoteInOperandsInQuery): Enables remoting in operands.
        "query_remote_in_operands": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionProgressiveQueryMinRowCountPerUpdate): Hint for Kusto as to how many records to send in each update 
        # (Takes effect only if OptionProgressiveQueryIsProgressive is set)
        "query_results_progressive_row_count": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # (OptionProgressiveProgressReportPeriod): Hint for Kusto as to how often to send progress frames (Takes effect only if OptionProgressiveQueryIsProgressive is set)
        "query_results_progressive_update_period": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # (OptionTakeMaxRecords): Enables limiting query results to this number of records. [Long]
        "query_take_max_records": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # (OptionQueryConsistency): Controls query consistency. ['strongconsistency' or 'normalconsistency' or 'weakconsistency']
        "queryconsistency": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "enum", "values": ['strongconsistency', 'normalconsistency', 'weakconsistency']},

        # (OptionRequestBlockRowLevelSecurity): If specified, blocks access to tables for which row_level_security policy is enabled [Boolean]
        "request_block_row_level_security": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionRequestCalloutDisabled): If set, callouts to external services are blocked. [Boolean]
        "request_callout_disabled": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionRequestDescription): Arbitrary text that the author of the request wants to include as the request description. [String]
        "request_description": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},

        # (OptionRequestExternalTableDisabled): If specified, indicates that the request cannot invoke code in the ExternalTable. [bool]
        "request_external_table_disabled": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionDoNotImpersonate): If specified, indicates that the service shouldn't impersonate the caller's identity. [bool]
        "request_impersonation_disabled": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionRequestReadOnly): If specified, indicates that the request must not be able to write anything. [Boolean]
        "request_readonly": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionRequestEntitiesDisabled): If specified, indicates that the request cannot access remote databases and clusters. [bool]
        "request_remote_entities_disabled": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionRequestSandboxedExecutionDisabled): If specified, indicates that the request cannot invoke code in the sandbox. [bool]
        "request_sandboxed_execution_disabled": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # NOT DOCUMENTED - (OptionResponseDynamicSerialization): Controls the serialization of 'dynamic' values in result sets. ['string', 'json']
        "response_dynamic_serialization": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "enum", "values": ['string', 'json']},

        # NOT DOCUMENTED - (OptionResponseDynamicSerialization_2): Controls the serialization of 'dynamic' string and null values in result sets. ['legacy', 'current']
        "response_dynamic_serialization_2": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "enum", "values": ['legacy', 'current']},

        # (OptionResultsProgressiveEnabled): If set, enables the progressive query stream
        "results_progressive_enabled": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # NOT DOCUMENTED - (OptionSandboxedExecutionDisabled): If set, using sandboxes as part of query execution is disabled. [Boolean]
        "sandboxed_execution_disabled": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},

        # (OptionServerTimeout): Overrides the default request timeout. [TimeSpan]
        # is capped by 1hour
        "servertimeout": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "str"},

        # (OptionTruncationMaxRecords): Overrides the default maximum number of records a query is allowed to return to the caller (truncation). [Long]
        "truncationmaxrecords": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # (OptionTruncationMaxSize): Overrides the dfefault maximum data size a query is allowed to return to the caller (truncation). [Long]
        "truncationmaxsize": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "uint"},

        # (OptionValidatePermissions): Validates user's permissions to perform the query and doesn't run the query itself. [Boolean]
        "validate_permissions": {"schema": [Schema.AZURE_DATA_EXPLORER], "type": "bool"},


        # For either implicit or explicit cross-application queries, specify resources you will be accessing
        # see https://dev.loganalytics.io/documentation/Using-the-API/Cross-Resource-Queries
        "workspaces": {"schema": [Schema.LOG_ANALYTICS], "type": "list"},

        # For either implicit or explicit cross-application queries, specify resources you will be accessing
        # see: https://dev.applicationinsights.io/documentation/Using-the-API/Cross-Resource-Queries
        "applications": {"schema": [Schema.APPLICATION_INSIGHTS, Schema.AIMON], "type": "list"},

        # The timespan over which to query data. This is an ISO8601 time period value. This timespan is applied in addition to any that are specified in the query expression.
        # see: https://docs.microsoft.com/en-us/rest/api/application-insights/query/get
        "timespan": {"schema": [Schema.APPLICATION_INSIGHTS, Schema.AIMON, Schema.LOG_ANALYTICS], "type": "iso8601_duration"},
    }


    # all lookup keys in table, must be without spaces, underscores and hypthen-minus, because parser ignores them
    _OPTIONS_TABLE:Dict[str,Dict[str,Any]] = {
        "ad": {"abbreviation": "autodataframe"},
        "autodataframe": {"flag": "auto_dataframe", "type": "bool"},
        "se": {"abbreviation": "shorterrors"},
        "shorterrors": {"flag": "short_errors", "type": "bool"},
        "f": {"abbreviation": "feedback"},
        "feedback": {"flag": "feedback", "type": "bool"},
        "sci": {"abbreviation": "showconninfo"},
        "showconninfo": {"flag": "show_conn_info", "type": "str", "allow_none": True},
        "c2lv": {"abbreviation": "columnstolocalvars"},
        "columnstolocalvars": {"flag": "columns_to_local_vars", "type": "bool"},
        "sqt": {"abbreviation": "showquerytime"},
        "showquerytime": {"flag": "show_query_time", "type": "bool"},
        "sq": {"abbreviation": "showquery"},
        "showquery": {"flag": "show_query", "type": "bool"},
        "sql": {"abbreviation": "showquerylink"},
        "showquerylink": {"flag": "show_query_link", "type": "bool"},
        "qld": {"abbreviation": "querylinkdestination"},
        "querylinkdestination": {"flag": "query_link_destination", "type": "str"},

        "esr": {"abbreviation": "enablesuppressresult"},
        "enablesuppressresult": {"flag": "enable_suppress_result", "type": "bool"},
        "pfi": {"abbreviation": "plotlyfsincludejs"},
        "plotlyfsincludejs": {"flag": "plotly_fs_includejs", "type": "bool"},
        "pw": {"abbreviation": "popupwindow"},
        "popupwindow": {"flag": "popup_window", "type": "bool", "init": False},
        "al": {"abbreviation": "autolimit"},
        "autolimit": {"flag": "auto_limit", "type": "int", "allow_none": True},
        "dl": {"abbreviation": "displaylimit"},
        "displaylimit": {"flag": "display_limit", "type": "int", "allow_none": True},
        "wait": {"abbreviation": "timeout"},
        "to": {"abbreviation": "timeout"},
        "timeout": {"flag": "timeout", "type": "int", "allow_none": True},
        "ptst": {"abbreviation": "prettytablestyle"},
        "prettytablestyle": {"flag": "prettytable_style", "type": "str"},
        "var": {"abbreviation": "lastrawresultvar"},
        "lastrawresultvar": {"flag": "last_raw_result_var", "type": "str"},
        "tp": {"abbreviation": "tablepackage"},
        "tablepackage": {"flag": "table_package", "type": "str"},
        "pp": {"abbreviation": "plotpackage"},
        "plotpackage": {"flag": "plot_package", "type": "str"},
        "df": {"abbreviation": "dsnfilename"},
        "dsnfilename": {"flag": "dsn_filename", "type": "str", "allow_none": True},
        "vc": {"abbreviation": "validateconnectionstring"},
        "validateconnectionstring": {"flag": "validate_connection_string", "type": "bool"},
        "aps": {"abbreviation": "autopopupschema"},
        "autopopupschema": {"flag": "auto_popup_schema", "type": "bool"},
        "jd": {"abbreviation": "jsondisplay"},
        "jsondisplay": {"flag": "json_display", "type": "str"},
        "sjd": {"abbreviation": "schemajsondisplay"},
        "schemajsondisplay": {"flag": "schema_json_display", "type": "str"},
        "pd": {"abbreviation": "palettedesaturation"},
        "palettedesaturation": {"flag": "palette_desaturation", "type": "float"},
        "pn": {"abbreviation": "palettename"},
        "paramsdict": {"flag": "params_dict", "type": "dict", "init": None},
        "palettename": {"flag": "palette_name", "type": "str"},
        "cache": {"flag": "cache", "type": "str", "allow_none": True},
        "usecache": {"flag": "use_cache", "type": "str", "allow_none": True},
        
        "tempfoldername": {"flag": "temp_folder_name", "type": "str"},
        "cachefoldername": {"flag": "cache_folder_name", "type": "str"},
        "exportfoldername": {"flag": "export_folder_name", "type": "str"},
        "addkqlreftohelp": {"flag": "add_kql_ref_to_help", "type": "bool"},
        "addschematohelp": {"flag": "add_schema_to_help", "type": "bool"},
        "notebookapp": {"flag": "notebook_app", "type": "str"},
        "debug": {"flag": "debug", "type": "bool"},

        "checkmagicversion": {"flag": "check_magic_version", "type": "bool"},
        "showwhatnew": {"flag": "show_what_new", "type": "bool"},
        "showinitbanner": {"flag": "show_init_banner", "type": "bool"},
        "iskernelintializtion": {"flag": "is_kernel_intializtion", "type": "bool"},
        "warnmissingdependencies": {"flag": "warn_missing_dependencies", "type": "bool"},
        "warnmissingenvvariables": {"flag": "warn_missing_env_variables", "type": "bool"},
        "allowsinglelinecell": {"flag": "allow_single_line_cell", "type": "bool"},
        "allowpycommentsbeforecell": {"flag": "allow_py_comments_before_cell", "type": "bool"},
        
        "kqlmagickernel": {"flag": "kqlmagic_kernel", "type": "bool"},

        "extrasrequire": {"flag": "extras_require", "type": "str"},

        "testnotebookapp": {"flag": "test_notebook_app", "type": "str"},

        "cloud": {"flag": "cloud", "type": "str"},
        "enablesso": {"flag": "enable_sso", "type": "bool"},
        "ssodbgcinterval": {"flag": "sso_db_gc_interval", "type": "int"},
        
        "authusehttpclient": {"flag": "auth_use_http_client", "type": "bool"},

        "tryazclilogin": {"flag": "try_azcli_login", "type": "bool"},
        "tryazcliloginbyprofile": {"flag": "try_azcli_login_by_profile", "type": "bool"},
        "tryvscodelogin": {"flag": "try_vscode_login", "type": "bool"},
        "tryazcliloginsubscription": {"flag": "try_azcli_login_subscription", "type": "str", "allow_none": True},
        "trytoken": {"flag": "try_token", "type": "dict", "allow_none": True},
        "trymsi": {"flag": "try_msi", "type": "dict", "allow_none": True},

        "idtag": {"abbreviation": "requestidtag"},
        "requestidtag": {"flag": "request_id_tag", "type": "str", "allow_none": True},

        "apptag": {"abbreviation": "requestapptag"},
        "requestapptag": {"flag": "request_app_tag", "type": "str", "allow_none": True},

        "usertag": {"abbreviation": "requestusertag"},
        "requestusertag": {"flag": "request_user_tag", "type": "str", "allow_none": True},
        
        "uatag": {"abbreviation": "requestuseragenttag"},
        "useragenttag": {"abbreviation": "requestuseragenttag"},
        "requestuatag": {"abbreviation": "requestuseragenttag"},
        "requestuseragenttag": {"flag": "request_user_agent_tag", "type": "str", "allow_none": True},

        "maxage": {"abbreviation": "requestcachemaxage"},
        "requestcachemaxage": {"flag": "request_cache_max_age", "type": "int", "allow_none": True},

        "dcln": {"abbreviation": "devicecodeloginnotification"},
        "devicecodeloginnotification": {"flag": "device_code_login_notification", "type": "str"},

        "dcne": {"abbreviation": "devicecodenotificationemail"},
        "devicecodenotificationemail": {"flag": "device_code_notification_email", "type": "str"},

        "saveas": {"flag": "save_as", "type": "str", "init": None},
        "saveto": {"flag": "save_to", "type": "str", "init": None},
        "query": {"flag": "query", "type": "str", "init": None},
        "conn": {"flag": "conn", "type": "str", "init": None},
        "queryproperties": {"flag": "query_properties", "type": "dict", "init": None},

        "pc": {"abbreviation": "palettecolors"},
        "palettecolors": {"flag": "palette_colors", "type": "int"},
        "pr": {"abbreviation": "palettereverse"},
        "palettereverse": {"flag": "palette_reverse", "type": "bool", "init": False},

        "ps": {"abbreviation": "popupschema"},
        "popupschema": {"flag": "popup_schema", "type": "bool", "init": False},

        "did": {"abbreviation": "displayid"},
        "displayid": {"flag": "display_id", "type": "bool", "init": False},
        "displayhandlers": {"flag": "display_handlers", "type": "dict", "init": {}},

        "pi": {"abbreviation": "popupinteraction"},        
        "popupinteraction": {"flag": "popup_interaction", "type": "str"},
        "tempfilesserver": {"flag": "temp_files_server", "type": "str"},
        "tempfilesserveraddress": {"flag": "temp_files_server_address", "type": "str", "allow_none": True},
        
        "kernellocation": {"flag": "kernel_location", "type": "str"},
        "kernelid": {"flag": "kernel_id", "type": "str", "allow_none": True},

        "notebookserviceaddress": {"flag": "notebook_service_address", "type": "str", "allow_none": True},

        "dtd": {"abbreviation": "dynamictodataframe"},
        "dynamictodataframe": {"flag": "dynamic_to_dataframe", "type": "str"},

        "tempfolderlocation": {"flag": "temp_folder_location", "type": "str"},

        "pl": {"abbreviation": "plotlylayout"},
        "plotlylayout": {"flag": "plotly_layout", "type": "dict", "allow_none": True},

        "atw": {"abbreviation": "authtokenwarnings"},
        "authtokenwarnings": {"flag": "auth_token_warnings", "type": "bool"},

        "ecbp": {"abbreviation": "enablecurlybracketsparams"},
        "enablecurlybracketsparams": {"flag": "enable_curly_brackets_params", "type": "bool"},

        "nop": {"flag": "nop", "type": "bool", "init": False},  # does nothing, useful to indicate option part when no options are required

        "av": {"abbreviation": "assignvar"},
        "assignvar": {"flag": "assign_var", "type": "str", "allow_none": True},
        "cv": {"abbreviation": "cursorvar"},
        "cursorvar": {"flag": "cursor_var", "type": "str", "allow_none": True},
        "ismagic": {"flag": "is_magic", "type": "bool"},

        "caim": {"abbreviation": "codeauthinteractivemode"},
        "codeauthinteractivemode": {"flag": "code_auth_interactive_mode", "type": "str", "allow_none": True},
    }



    @classmethod
    def validate_override(cls, name:str, config:Configurable, **override_options)->Dict[str,Any]:
        """validate the provided option are valid"""

        options = {}
        for key, value in override_options.items():
            obj = cls._get_obj(key, allow_abbr=True)
            if obj.get("flag") in config.read_only_trait_names:
                raise ValueError(f"option '{key}' in {name} is readony, cannot be set")
            cls._convert(name, obj, key, value)
            cls._validate_config_trait(name, obj, key, value, config)
            options[obj.get("flag")] = value

        return options


    @classmethod
    def parse_option(cls, dict_name:str, key:str, value:str, config:Configurable=None, lookup:Dict[str,Dict[str,Any]]=None, user_ns:Dict[str,Any]=None, allow_abbr:bool=None, force:bool=False):
        """validate the provided option are valid
           return normalized key and value"""

        obj = cls._get_obj(key, lookup=lookup, allow_abbr=allow_abbr)
        value = cls._parse_value(dict_name, obj, key, value, user_ns=user_ns)
        cls._validate_config_trait(dict_name, obj, key, value, config)
        key_name = obj.get("flag", key)
        if config.is_read_only(key_name) and value != getattr(config, key_name):
            # done to raise the proer error
            setattr(config, key_name, value)

        return key_name, value


    @classmethod
    def _parse_value(cls, dict_name:str, obj:Dict[str,Any], key:str, string:str, user_ns:Dict[str,Any])->Any:
        _type = obj.get("type")

        if string == "" and _type == "str":
            return string

        # if we allow to bring value from python, we also allow from env variables
        # when we parse env vironment with option we fon't use user_ns
        if string.startswith('$') and user_ns:
            env_var_name = string[1:]
            if not is_env_var(env_var_name):
                raise ValueError(f"failed to parse referred value, due environment variable {env_var_name} not set")
            string = get_env_var(env_var_name)
            _was_quoted, value = strip_if_quoted(string)
        else:
            try:
                value = eval(string, None, user_ns)
            except:
                # if no user_ns it means parse is for environment var, and it just may be an unquoted object
                if user_ns:
                    raise
                value = string

        # check value is of the right type
        try:
            return cls._convert(dict_name, obj, key, value)
        except:
            raise


    @classmethod
    def parse_config_key(cls, key:str, config:Configurable, allow_abbr:bool=None)->Tuple[str,str,Any]:
        """validate the provided option key is valid
           return normalized key"""

        obj = cls._get_obj(key, allow_abbr=allow_abbr) 
        name = obj.get("flag")
        if "init" in obj:
            value = obj.get("init")
        elif name in cls.traits_dict:
            value = getattr(config, name)
        else:
            raise f"internal error '{key}' has no init value and not defined as Kqlmagic traitlet"
        return name, value


    @classmethod
    def _get_obj(cls, key:str, lookup:Dict[str,Dict[str,Any]]=None, allow_abbr:bool=None)->Dict[str,Any]:
        lookup_key = key.lower().replace("-", "").replace("_", "")
        lookup_table = lookup or cls._OPTIONS_TABLE
        obj = lookup_table.get(lookup_key)
        if obj is None:
            raise ValueError(f"unknown option '{key}'")

        if obj.get("abbreviation"):
            obj = lookup_table.get(obj.get("abbreviation"))
            if not allow_abbr is not True:
                raise ValueError(f"unknown option '{key}'. (Found option abbreviation '{key}' for {obj.get('flag')})")
        return obj


    @classmethod
    def _parse_kql_options(cls, code:str, is_cell:bool, config:Configurable, user_ns:Dict[str,Any])->Tuple[str,Dict[str,Any]]:
        trimmed_kql = code
        trimmed_kql = trimmed_kql.strip()
        suppress_results = False
        if trimmed_kql.endswith(";"):
            suppress_results = not is_cell
            if is_cell:
                lines = trimmed_kql.splitlines(True)
                if lines[-1].strip() == ";":
                    suppress_results = True
            if suppress_results:
                trimmed_kql = trimmed_kql[:-1].strip()
        
        words = trimmed_kql.split()

        properties = {}

        table = options = cls.default_options.copy()

        if not words:
            return ("", options)
        num_words = len(words)
        first_word = 0

        if num_words - first_word >= 2 and words[first_word + 1] == "<<":
            options["result_var"] = words[first_word]
            trimmed_kql = trimmed_kql[trimmed_kql.find("<<") + 2:]
            first_word += 2

        obj = None
        key = None
        opt_key = None
        key_state = True
        option_type = None
        is_option = True
        is_property = False
        skip_words_count = 0
        for word in words[first_word:]:
            if key_state:

                if skip_words_count == 0:
                    _comment, skip_words_count = cls._parse_comment(word, trimmed_kql)              

                if skip_words_count > 0:
                    skip_words_count -= 1
                    trimmed_kql = trimmed_kql[trimmed_kql.find(word) + len(word):]
                    continue

                is_option = word.startswith("-")
                is_property = word.startswith("+")
                option_type = "option" if is_option else "query property"
                if not is_option and not is_property:
                    break
                # validate it is not a command
                if is_option and word.startswith("--"):
                    raise ValueError(f"invalid {option_type} '{word}', cannot start with a bouble hyphen-minus")

                trimmed_kql = trimmed_kql[trimmed_kql.find(word) + len(word):]
                word = word[1:]
                bool_value = True
                if word[0].startswith("!"):
                    bool_value = False
                    word = word[1:]
                if "=" in word:
                    parts = word.split("=", 1)
                    key = parts[0]
                    value = parts[1]
                else:
                    key = word
                    value = None

                if is_option:
                    lookup_key = key.lower().replace("-", "").replace("_", "")
                    obj = cls._OPTIONS_TABLE.get(lookup_key)
                    table = options 
                else:
                    lookup_key = key.lower()
                    obj = cls._QUERY_PROPERTIES_TABLE.get(lookup_key)
                    table = properties 

                if obj is not None:
                    if obj.get("abbreviation") is not None:
                        obj = cls._OPTIONS_TABLE.get(obj.get("abbreviation"))
                    if obj.get("flag") in config.read_only_trait_names:
                        raise ValueError(f"{option_type} {key} is readony, cannot be set")

                    _type = obj.get("type")
                    opt_key = obj.get("flag") or lookup_key
                    if _type == "bool" and value is None:
                        table[opt_key] = bool_value
                    else:
                        if not bool_value:
                            raise ValueError(f"{option_type} {key} cannot be negated")
                        if value is not None:
                            table[opt_key] = cls._parse_value("options" if is_option else "query properties", obj, key, value, user_ns)
                        else:
                            key_state = False
                else:
                    raise ValueError(f"unknown {option_type} '{key}'")
            else:
                trimmed_kql = trimmed_kql[trimmed_kql.find(word) + len(word):]
                table[opt_key] = cls._parse_value("options", obj, key, word, user_ns)
                key_state = True
            first_word += 1

            # validate using config traits
            if key_state and is_option:
                cls._validate_config_trait("options", obj, key, options.get(opt_key), config)
            
        if not key_state:
            raise ValueError(f"{option_type} '{opt_key}' must have a value")

        if options.get("query_properties"):
            properties.update(options["query_properties"])
        options["query_properties"] = properties
        if suppress_results:
            options["suppress_results"] = True
        return (trimmed_kql.strip(), options)


    @classmethod
    def _parse_comment(cls, word:str, _str:str)->Tuple[str,int]:

        comment = None
        skip_words_count = 0
        if word.startswith("//"):
            idx_start = _str.find(word)
            idx_end = _str[idx_start:].find("\n")
            if idx_end > 0:
                idx_end = idx_start + idx_end
                comment = _str[idx_start:idx_end]
            else:
                comment = _str[idx_start:]
            comment_words = comment.split()
            skip_words_count = len(comment_words)

        return comment, skip_words_count


    @classmethod
    def parse_and_get_kv_string(cls, conn_str:str, user_ns:Dict[str,Any], keep_original_key:bool=None)->Dict[str,Any]:
        rest = conn_str
        rest = rest.strip()
        _was_quoted, rest = strip_if_quoted(rest)
            
        matched_kv = {}

        delimiter_required = False
        lp_idx = rest.find("(") 
        eq_idx = rest.find("=") 
        sc_idx = rest.find(";") 
        l_char = "(" if eq_idx < 0 and sc_idx < 0 else "=" if lp_idx < 0 else "(" if lp_idx < eq_idx and lp_idx < sc_idx else "="
        r_char = ")" if l_char == "(" else ";"
        extra_delimiter = None if r_char == ";" else "."

        while len(rest) > 0:
            l_idx = rest.find(l_char)
            r_idx = rest.find(r_char)
            if l_idx < 0:
                if l_char == "(":
                    # string ends with delimiter
                    if extra_delimiter is not None and extra_delimiter == rest:
                        break
                    else:
                        raise ValueError("invalid key/value string, missing left parethesis.")

                # key only at end of string
                elif r_idx < 0:
                    key = rest
                    val = ""
                    rest = ""

                # key only
                else:
                    key = rest[:r_idx].strip()
                    val = ""
                    rest = rest[r_idx + 1:].strip()

            # key only
            elif r_idx >= 0 and r_idx < l_idx:
                if l_char == "(":
                    raise ValueError("invalid key/value string, missing left parethesis.")
                else:
                    key = rest[:r_idx].strip()
                    val = ""
                    rest = rest[r_idx + 1:].strip()

            # key and value
            else:
                key = rest[:l_idx].strip()
                rest = rest[l_idx + 1:].strip()
                r_idx = rest.find(r_char)
                if r_idx < 0:
                    if l_char == "(":
                        raise ValueError("invalid key/value string, missing right parethesis.")
                    else:
                        val = rest
                        rest = ""
                else:
                    val = rest[:r_idx].strip()
                    rest = rest[r_idx + 1:].strip()
                if extra_delimiter is not None:
                    if key.startswith(extra_delimiter):
                        key = key[1:].strip()
                    elif delimiter_required:
                        raise ValueError("invalid key/value string, missing delimiter.")
                    delimiter_required = True

            # key exist
            if len(key) > 0:
                if keep_original_key is True:
                    lookup_key = key
                else:
                    val = cls._parse_value("key/value", {"type": "str"}, key, val, user_ns)
                    lookup_key = key.lower().replace("-", "").replace("_", "")
                matched_kv[lookup_key] = val

            # no key but value exist
            elif len(val) > 0:
                raise ValueError("invalid key/value string, missing key.")

            # no key, no value in parenthesis mode
            elif l_char == "(":
                raise ValueError("invalid key/value string, missing key.")

        return matched_kv


    @classmethod
    def parse_quote(cls, string:str):
        string = string.strip()
        delimiter = string[0]
        delimiter_len = 1
        triple_quote = len(string) > 2 and string[1] == delimiter and string[2] == delimiter
        if triple_quote:
            delimiter_len = 3
            delimiter = delimiter * 3
            quoted_string_len = string[3:].find(delimiter)

        else:
            escape = False
            quoted_string_len = -1

            count = 0
            for c in string[1:]:
                if c == "\\":
                    escape = not escape
                elif escape:
                    pass
                elif c == delimiter:
                    quoted_string_len = count
                    break
                count += 1

        if quoted_string_len >= 0:
            trimmed_string = string[quoted_string_len + 2 * delimiter_len:]
            if len(trimmed_string) > 0 and not trimmed_string[0].isspace():
                raise SyntaxError("invalid syntax after quoted string, should be followed by whitespace only")

            quoted_string = string[delimiter_len:quoted_string_len + delimiter_len]
            quoted_words = len(quoted_string.split())
            if len(quoted_string)> 0:
                if quoted_string[-1].isspace():
                    quoted_words += 1
                if quoted_string[0].isspace():
                    quoted_words += 1
            else:
                quoted_words = 1
            return delimiter + quoted_string + delimiter, quoted_words
        else:
            raise SyntaxError("EOL while scanning quoted string")


    @classmethod
    def _convert(cls, name:str, obj:Dict[str,Any], key:str, value:Any)->Any:
        if value is None:
            if obj.get("allow_none"):
                return None
            else:
                raise ValueError(f"option '{key}' doesn't allow None value.")
        _type = None
        try:
            _type = obj.get("type")
            if _type == "int":
                if float(value) != int(value):
                    raise ValueError
                return int(value)
            elif _type == "uint":
                if float(value) != int(value) or int(value) < 0:
                    raise ValueError
                return int(value)                 
            elif _type == "float":
                return float(value)
            elif _type == "bool":
                if type(value) == str:
                    if value.lower() == 'true':
                        return True
                    elif value.lower() == 'false':
                        return False
                    else:
                        raise ValueError
                elif bool(value) != int(value):
                    raise ValueError
                return bool(value)
            elif _type == "dict":
                return dict(value)
            elif _type == "list":
                if type(value) == str:
                    value = [value]                     
                return list(value)
            elif _type == "enum":
                enum_values = obj.get("values", [])
                if enum_values.index(value) >= 0:
                    return value
                else:
                    raise ValueError
            elif _type == "iso8601_duration":
                # There are four ways to express a time interval:
                # Start and end, such as "2007-03-01T13:00:00Z/2008-05-11T15:30:00Z"
                # Start and duration, such as "2007-03-01T13:00:00Z/P1Y2M10DT2H30M"
                # Duration and end, such as "P1Y2M10DT2H30M/2008-05-11T15:30:00Z"
                # Duration only, such as "P1Y2M10DT2H30M", with additional context information

                value_list = [value] if type(value) != list else list(value)[:2]
                if len(value_list) == 0:
                    raise ValueError

                elif len(value_list) == 1:
                    value = value_list[0]
                    if isinstance(value, timedelta):
                        isodate = Dependencies.get_module("isodate", message="timedelta convertion to iso8601 duration format is not supported without isodate module, use instead a datetime range format, or already converted string") # will throw if does not exist
                        value = isodate.duration_isoformat(value)
                    elif type(value) != str:
                        raise ValueError
                    return value

                else:
                    start_value = value_list[0]
                    end_value = value_list[1]
                    if isinstance(start_value, timedelta):
                        isodate = Dependencies.get_module("isodate", dont_throw=True)
                        if isodate:
                           start_value = isodate.duration_isoformat(start_value) 
                        else:
                            end_datetime = end_value if isinstance(end_value, datetime) else dateutil.parser.isoparse(end_value)
                            start_value = end_datetime - start_value
                    elif isinstance(end_value, timedelta):
                        isodate = Dependencies.get_module("isodate", dont_throw=True)
                        if isodate:
                           end_value = isodate.duration_isoformat(end_value)
                        else:
                            start_datetime = end_value if isinstance(start_value, datetime) else dateutil.parser.isoparse(start_value)
                            end_value = end_value + start_datetime
                    value_list = [v.strftime('%Y-%m-%dT%H:%M:%S%ZZ') if isinstance(v, datetime) else str(v) for v in [start_value, end_value]]
                    return "/".join(value_list)

            else:
                return str(value)

        except Exception as e:
            option_type = "property" if name == "query properties" else "option"
            due_message = f"{e}" or f"invalid '{_type}' of value '{value}'"
            raise ValueError(f"failed to set {option_type} '{key}' in {name}, due to {due_message}")


    @classmethod
    def _validate_config_trait(cls, dict_name:str, obj:Dict[str,Any], key:str, value:Any, config:Configurable)->None:           
        # validate using config traits
        name = obj.get("flag")
        if isinstance(config, Configurable) and name in cls.traits_dict:
            #
            # save current value
            #
            try:
                new_value = cls._convert(dict_name, obj, key, value)
                trait:TraitType = cls.traits_dict.get(name)
                if hasattr(trait, "_validate"):
                    validated_value = trait._validate(config, new_value)
                return
            except Exception as error:
                raise ValueError(f"failed to set option '{key}' in {dict_name}, due to invalid value '{value}'. Exception: {error}")



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

"""Simple HTTP Client"""
import json
import gzip
from typing import Dict, Any, Union

from urllib.parse import urlencode

from .dependencies import Dependencies
from .constants import Constants
from .log import logger
from ._version import __version__


requests_module = Dependencies.get_module("requests", dont_throw=True)

class HTTPErrorResponse(object):

    global requests_module
    if requests_module is None:
        from urllib.error import HTTPError
        def __init__(self, http_error: HTTPError)-> None:
            self._http_error = http_error
            self._text = self._http_error.read() if hasattr(self._http_error, "read") and callable(self._http_error.read) else ""
            self._code = self._http_error.code if hasattr(self._http_error, "code") else 500
            self._headers = self._http_error.headers if hasattr(self._http_error, "headers") else {}


        def getcode(self)-> int:
            return self._code


        def info(self)-> dict:
            return self._headers


        def read(self)-> bytes:
            return self._text


class Response(object):
    """This describes a minimal http response interface used by this package.
    :var int status_code:
        The status code of this http response.
        Our async code path would also accept an alias as "status".
    :var string text:
        The body of this http response.
        Our async code path would also accept an awaitable with the same name.
    """

    global requests_module
    if requests_module is None:
        from http.client import HTTPResponse

        def __init__(self, url: str, response: HTTPResponse)-> None:
            """
            :param response: The return value from a open call
                            on a urllib.build_opener()
            :type response:  urllib response object
            """
            self._url: str = url
            self._response = response
            self._status_code: int = None
            self._headers: dict = None
            self._content: bytes = None
            self._text: str = None


        @property
        def status_code(self)-> int:
            """status code"""
            self._status_code = self._status_code or self._response.getcode()
            return self._status_code


        @property
        def headers(self)-> dict:
            """response headers"""
            self._headers = self._headers or self._response.info()
            return self._headers


        @property
        def text(self)-> str:
            """response data string"""
            self._text = self._text or self._decode(self.content)
            return self._text


        @property
        def content(self)-> bytes:
            """response data content"""
            self._content = self._content or self._get_pagedata_bytes(self._response.read())
            return self._content

        
        @property
        def url(self)-> str:
            """url string"""
            return self._url


        @property
        def reason(self)-> bytes:
            """response error reason"""
            if 400 <= self.status_code:
                return self.content
            else:
                return None


        def json(self)-> Any:
            """response data object"""
            return json.loads(self.text)


        def raise_for_status(self)-> None:
            """Raises :class:`HTTPError`, if one occurred."""
            if 400 <= self.status_code:
                from urllib.error import HTTPError
                reason_text = self._get_reason_text()
                http_error_msg = f'{self.status_code} {"Client" if self.status_code < 500 else "Server"} Error: {reason_text} for url: {self.url}'
                raise HTTPError(self.url, self.status_code, http_error_msg, self.headers, None)

        
        def _get_pagedata_bytes(self, _bytes:bytes)-> str:
            if self.headers.get('Content-Encoding') == 'gzip':
                pagedata = gzip.decompress(_bytes)
            elif self.headers.get('Content-Encoding') == 'deflate':
                pagedata = _bytes
            elif self.headers.get('Content-Encoding'):
                print('Encoding type unknown')
                pagedata = _bytes
            else:
                pagedata = _bytes
            return pagedata


        def _get_reason_text(self)-> str:
            if isinstance(self.reason, bytes):
                # We attempt to decode utf-8 first because some servers
                # choose to localize their reason strings. If the string
                # isn't utf-8, we fall back to iso-8859-1 for all other
                # encodings. (See PR #3538)
                text = self._decode(self.reason)
            else:
                text = self.reason
            return text


        def _decode(self, _bytes: bytes)-> str:
            try:
                text = _bytes.decode('utf-8')
            except UnicodeDecodeError:
                text = _bytes.decode('iso-8859-1')
            return text


class HttpClient(object):
    """Simple http client based on urllib"""

    supported_methods = {'DELETE', 'GET', 'PATCH', 'POST', 'PUT', 'HEAD', 'OPTIONS'}

    def __init__(self, host: str=None, path: str=None, headers: dict=None, timeout: float=None)-> None:

        global requests_module
        self._requests = requests_module
        self._headers: dict = headers or self._default_headers()
        self._timeout: float = timeout
        self._url: str = host
        if path:
            self._url = f"{self._url}/{path}"


    def _default_headers(self)-> dict:
        return {
                # 'User-Agent': f'{Constants.MAGIC_PACKAGE_NAME}/{__version__}',
                # 'Accept-Encoding': 'gzip, deflate',
                # 'Accept': '*/*',
                # 'Connection': 'keep-alive',
            }


    def post(self, url: str, params: Union[dict,str]=None, data: Union[str,bytes,Any]=None, headers: dict=None, json: Any=None, timeout: float=None, **kwargs)-> Response:
        """HTTP post.
        :param string url: target url
        :param dict params: A dict to be url-encoded and sent as query-string.
        :param dict headers: A dict representing headers to be sent via request.
        :param data:
            Implementation needs to support 2 types.
            * An object, which will need to be urlencode() before being sent.
            * (Recommended) A string, which will be sent in request as-is.
        :param json: object, which will need to be urlencode() before being sent.
        It returns an :class:`~Response`-like object.
        Note: In its async counterpart, this method would be defined as async.
        """

        response = self._execute_http_request(url=url, method="POST", params=params, headers=headers, timeout=timeout, data=data, json=json, **kwargs)
        return response


    def get(self, url: str, params: Union[dict,str]=None, headers: dict=None, timeout: float=None, **kwargs)-> Response:
        """HTTP get.
        :param string url: target url
        :param dict params: A dict to be url-encoded and sent as query-string.
        :param dict headers: A dict representing headers to be sent via request.
        It returns an :class:`~Response`-like object.
        Note: In its async counterpart, this method would be defined as async.
        """

        response = self._execute_http_request(url=url, method="GET", params=params, headers=headers, timeout=timeout, **kwargs)
        return response


    def request(self, method: str=None, url: str=None, params: Union[dict,str]=None, headers: dict=None, timeout: float=None, data: Union[str,bytes,Any]=None, json: Any=None, **kwargs)-> Response:
        """HTTP request.
        :param string url: target 
        :param string method: http verb
        :param dict params: A dict to be url-encoded and sent as query-string.
        :param dict headers: A dict representing headers to be sent via request.
        :param data:
            Implementation needs to support 2 types.
            * An object, which will need to be urlencode() before being sent.
            * (Recommended) A string, which will be sent in request as-is.
        :param json: object, which will need to be urlencode() before being sent.
        It returns an :class:`~Response`-like object.
        Note: In its async counterpart, this method would be defined as async.
        """

        response = self._execute_http_request(url=url, method=method, params=params, headers=headers, timeout=timeout, data=data, json=json, **kwargs)
        return response


    def close(self):  # Not required, but we use it to avoid a warning in unit test
        pass


    def _execute_http_request(self, url: str=None, method: str=None, params: Union[dict,str]=None, headers: dict=None, timeout: float=None, data: Union[str,bytes,Any]=None, json: Any=None, **kwargs)-> Response:
        # assert not kwargs, "Our stack shouldn't leak extra kwargs: %s" % kwargs
        try:
            method = self._get_method(method=method, data=data or json)
            headers = self._get_headers(headers=headers, json=json)
            timeout = timeout or self._timeout

            if self._requests:
                url = self._build_url(url=url)
                logger().debug(f"{method} Request: {url}, params: {params}, payload: {data or json}, headers: {headers}")
                data: bytes = self._get_data(data=data)
                response = self._requests.request(method, url, params=params, headers=headers, timeout=timeout, data=data, json=json)
            else:
                import urllib.request
                from urllib.request import Request
                from urllib.error import HTTPError
                ssl_context = None
                if url.lower().startswith("https"):
                    import ssl
                    ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
                    # ssl_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
                    # ssl_context.verify_mode = ssl.CERT_REQUIRED
                    # ssl_context.check_hostname = True
                url = self._build_url(url=url, params=params)
                data: bytes = self._get_data(data=data, json=json)
                request = Request(url, data=data, headers=headers, method=method)
                logger().debug(f"{method} Request: {request.get_full_url()}, payload: {request.data}, headers: {request.headers}")

                try:
                    result = urllib.request.urlopen(request, timeout=timeout, context=ssl_context)
                except HTTPError as err:
                    result = HTTPErrorResponse(err)

                response = Response(url, result)

            logger().debug(f"{method} Response: {response.status_code} {response.content}")
            return response

        except Exception as err:
            code = err.code if hasattr(err, "code") else None
            reason = err.reason if hasattr(err, "reason") else err.read() if hasattr(err, "read") else None
            logger().debug(f"{method} {url} Response: {code} {reason}")
            raise err        


    def _build_url(self, url: str=None, params: Union[dict,str]=None)-> str:
        url = url or self._url
        if params:
            if type(params) == str:
                url_values = params
            elif isinstance(params, dict):
                url_values = urlencode(sorted(params.items()), True)
            else: 
                raise Exception(f"http_client support str|dict params only. type(prams) == {type(params)}")
            url = f"{url}?{url_values}"
        return url


    def _get_method(self, method: str=None, data: Union[str,bytes,Any]=None)-> str:
        method = method or ("GET" if data is None else "POST")
        method = method.upper()
        if method not in self.supported_methods:
            raise Exception(f"method {method} is not supported")
        return method


    def _get_headers(self, headers: dict=None, json: Any=None)-> dict:
        request_headers: dict = self._headers.copy()
        if headers:
            request_headers.update(headers)
            # Remove keys that are set to None.
            none_keys = [k for (k, v) in request_headers.items() if v is None]
            for key in none_keys:
                del request_headers[key]
        if json and 'Content-Type' not in request_headers:
            request_headers['Content-Type'] = 'application/json'
        return request_headers


    def _get_data(self, data: Union[str,bytes,Any]=None, json: Any=None)-> bytes:
        if data is not None:
            if type(data) != bytes:
                if type(data) == str:
                    data = data.encode('utf-8')
                elif type(data) == dict:
                    data = urlencode(data).encode('utf-8')
                else:
                    data = self._json_dumps_encode(data)
        elif json is not None:
            data = self._json_dumps_encode(json)
        return data


    def _json_dumps_encode(self, data: Any)-> bytes:
        data = json.dumps(data, allow_nan=False).encode('utf-8')
        return data



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from typing import Any, Dict, List


# must be one of the fist to be executed, as it contains the information what is installed
from .dependencies import Dependencies 
dependencies = Dependencies()


from .log import logger

logger().debug("kql_magic.py - import Configurable from traitlets.config.configurable")
from traitlets.config.configurable import Configurable
logger().debug("kql_magic.py - import Bool, Int, Float, Unicode, Enum, TraitError, validate from traitlets")
from traitlets import TraitType, Bool, Int, Float, Unicode, Enum, TraitError, validate
from traitlets import Dict as _Dict

from .ipython_api import Magics, magics_class, cell_magic, line_magic, needs_local_scope, is_magics_class
from .ipython_api import IPythonAPI

from .kql_magic_core import Kqlmagic_core
from .constants import Constants, Cloud
from .palette import Palettes, Palette
from .os_dependent_api import OsDependentAPI



try:
    from flask import Flask
except Exception:
    flask_installed = False
else:
    flask_installed = True


from .results import ResultSet

kql_core_count:int = 0
kql_core_obj:Kqlmagic_core = None
is_non_magic_kql_on:bool = False


@magics_class
class Kqlmagic(Magics, Configurable):

    is_ipython_extension:bool = False
    is_kqlmagic_kernel:bool = False

    is_kernel_intializtion = Bool(
        default_value=False,
        read_only=True,
        config=True, 
        help="""when set, kernel is in init mode.\n"""
    )

    is_magic = Bool(
        default_value=False,
        read_only=True,
        config=True, 
        help="""when set, module is used as ipython magic.\n"""
    )

    auto_limit = Int(
        default_value=0, 
        config=True, 
        allow_none=True, 
        help="""Automatically limit the size of the returned result sets.\n
        Abbreviation: 'al'"""
    )

    prettytable_style = Enum(
        ["DEFAULT", "MSWORD_FRIENDLY", "PLAIN_COLUMNS", "RANDOM"],
        default_value="DEFAULT",
        config=True,
        help="""Set the table printing style to any of prettytable's defined styles.\n
        Abbreviation: 'ptst'"""
    )

    short_errors = Bool(
        default_value=True, 
        config=True, 
        help="""Don't display the full traceback on KQL Programming Error.\n
        Abbreviation: 'se'"""
    )

    display_limit = Int(
        default_value=None,
        config=True,
        allow_none=True,
        help="""Automatically limit the number of rows displayed (full result set is still stored).\n
        Abbreviation: 'dl'""",
    )

    auto_dataframe = Bool(
        default_value=False, 
        config=Dependencies.is_installed("pandas"), 
        help="""Return Pandas dataframe instead of regular result sets.\n
        Abbreviation: 'ad'"""
    )

    columns_to_local_vars = Bool(
        default_value=False, 
        config=True, 
        help="""Return data into local variables from column names.\n
        Abbreviation: 'c2lv'"""
    )

    feedback = Bool(
        default_value=True, 
        config=True, 
        help="""Show number of records returned, and assigned variables.\n
        Abbreviation: 'f'"""
    )

    show_conn_info = Enum(
        ["list", "current", "None"],
        default_value="current",
        config=True,
        allow_none=True,
        help="""Show connection info, either current, the whole list, or None.\n
        Abbreviation: 'sci'"""
    )

    dsn_filename = Unicode(
        default_value="odbc.ini",
        config=True,
        allow_none=True,
        help="""Sets path to DSN file.\n
        When the first argument of the connection string is of the form [section], a kql connection string is formed from the matching section in the DSN file.\n
        Abbreviation: 'dl'"""
    )

    cloud = Enum(
        [Cloud.PUBLIC, Cloud.MOONCAKE, Cloud.CHINA, Cloud.FAIRFAX, Cloud.GOVERNMENT, Cloud.BLACKFOREST, Cloud.GERMANY, Cloud.PPE],
        default_value=Cloud.PUBLIC,
        config=True,
        help="""Default cloud\n
        the kql connection will use the cloud as specified"""
    )

    enable_sso = Bool(
        default_value=False, 
        config=True, 
        help=f"""Enables or disables SSO.\n
        If enabled, SSO will only work if the environment parameter {Constants.MAGIC_CLASS_NAME_UPPER}_SSO_ENCRYPTION_KEYS is set properly."""
    )

    try_kqlmaic_sso = Bool(
        default_value=False, 
        config=True, 
        help=f"""Try to get token from Kqlmagic."""
    )

    try_azcli_login = Bool(
        default_value=False, 
        config=True, 
        help=f"""Try first to get token from Azure CLI.\n
        To override default tenant, tenant=value should be specified in connection string"""
    )

    try_azcli_login_by_profile = Bool(
        default_value=False, 
        config=True,
        help=f"""Try first to get token from Azure CLI by profile.\n
        To override default tenant, tenant=value should be specified in connection string."""
    )

    try_azcli_login_subscription = Unicode(
        default_value=None, 
        allow_none=True,
        config=True, 
        help=f"""Try first to get token from Azure CLI, for the specified subscription.\n
        To override default tenant, tenant=value should be specified in connection string."""
    )

    try_vscode_login = Bool(
        default_value=False, 
        config=True, 
        help=f"""Try first to get token from Visual Studio Code Azure Account login.\n
        To override default tenant, tenant=value should be specified in connection string."""
    )

    try_token = _Dict(
        default_value=None,
        config=True, 
        allow_none=True, 
        help=f"""Try first to use this token.\n
        Should be a dictionary with at least this keys: tokenType/token_type, accessToken/access_token"""
    )

    try_msi = _Dict(
        default_value=None,
        config=True, 
        allow_none=True, 
        help=f"""Try first to get MSI token from MSI endpoint.
        Should be a dictionary with the optional MSI params: resource, client_id/object_id/mis_res_id, cloud_environment, , timeout"""  
        # - timeout: If provided, must be in seconds and indicates the maximum time we'll try to get a token before raising MSIAuthenticationTimeout
        # - client_id: Identifies, by Azure AD client id, a specific explicit identity to use when authenticating to Azure AD. Mutually exclusive with object_id and msi_res_id.
        # - object_id: Identifies, by Azure AD object id, a specific explicit identity to use when authenticating to Azure AD. Mutually exclusive with client_id and msi_res_id.
        # - msi_res_id: Identifies, by ARM resource id, a specific explicit identity to use when authenticating to Azure AD. Mutually exclusive with client_id and object_id.
        # - cloud_environment (msrestazure.azure_cloud.Cloud): A targeted cloud environment
        # - resource (str): Alternative authentication resource, default is 'https://management.core.windows.net/'.      
    )

    sso_db_gc_interval = Int(
        default_value=168, 
        config=True,
        help="""Garbage Collection interval for not changed SSO cache entries. Default is one week."""
    )

    device_code_login_notification = Enum(
        ["auto", "button", "popup_interaction", "browser", "browser_reference", "terminal", "terminal_reference", "email"],
        default_value="auto", 
        config=True,
        help="""Sets device_code login notification method.\n
        Abbreviation: 'dcln'"""
    )

    device_code_notification_email = Unicode(
        default_value="", 
        config=True, 
        help=f"""Email details. Should be set by {Constants.MAGIC_CLASS_NAME_UPPER}_DEVICE_CODE_NOTIFICATION_EMAIL.\n
        the email details string format is: SMTPEndPoint='endpoint';SMTPPort='port';sendFrom='email';sendFromPassword='password';sendTo='email';context='text'\n
        Abbreviation: 'dcne'"""
    )

    # set by default to "device_code" and not "auto" for backward compitability
    code_auth_interactive_mode = Enum(
        ["auto", "device_code", "auth_code"],
        default_value="device_code",
        config=True,
        allow_none=True,
        help="""Sets code authentication interative mode, if "auto" is set, "auth_code" will be seleted if kernel is local, otherwise "device_code" is used\n
        Abbreviation: 'caim'"""
    )

    timeout = Int(
        default_value=None, 
        config=True, 
        allow_none=True, 
        help="""Specifies the maximum time in seconds, to wait for a query response. None, means default http wait time.\n
        Abbreviation: 'to' or 'wait'"""
    )

    enum_list = ["None"] 
    default_value = "None"
    if dependencies.is_installed("plotly"):
        default_value = "plotly"
        enum_list.append("plotly")
        enum_list.append("plotly_widget")
        enum_list.append("plotly_orca")
    plot_package = Enum(
        enum_list, 
        default_value=default_value, 
        config=len(enum_list) > 1, 
        help="""Set the plot package (plotlt_orca requires plotly orca to be installed on the server).\n 
        Abbreviation: 'pp'"""
    )

    enum_list = ["auto", "prettytable", "qgrid"] 
    if dependencies.is_installed("pandas"):
        enum_list.append("pandas")
        enum_list.append("pandas_html_table_schema")
    if dependencies.is_installed("plotly"):
        enum_list.append("plotly")
    table_package = Enum(
        enum_list, 
        default_value="auto", 
        config=True, 
        help="Set the table display package. Abbreviation: tp"
    )

    assign_var = Unicode(
        default_value=None,
        allow_none=True, 
        config=True, 
        help="""If specified, the query result result will be assigned to this variable in user's namespace.\n
        Abbreviation: 'av'"""
    )

    cursor_var = Unicode(
        default_value=None,
        allow_none=True, 
        config=True, 
        help="""If specified, the cursor value returned from the query will be assigned to this variable in user's namespace.\n
        Abbreviation: 'ac'"""
    )

    last_raw_result_var = Unicode(
        default_value="_kql_raw_result_", 
        config=True, 
        help="""Set the name of the variable that will contain last raw result.\n
        Abbreviation: 'var'"""
    )

    enable_suppress_result = Bool(
        default_value=True, 
        config=True, 
        help="""Suppress result when magic ends with a semicolon ;.\n 
        Abbreviation: 'esr'"""
    )

    show_query_time = Bool(
        default_value=True, 
        config=True, 
        help="""Print query execution elapsed time.\n 
        Abbreviation: 'sqt'"""
    )

    show_query = Bool(
        default_value=False, 
        config=True, 
        help="""Print parametrized query.\n
        Abbreviation: 'sq'"""
    )

    show_query_link = Bool(
        default_value=False, 
        config=True, 
        help="""Show query deep link as a button, to run query in the deafult tool.\n
        Abbreviation: ''sql'"""
    )

    query_link_destination = Enum(
        ["Kusto.Explorer", "Kusto.WebExplorer"], 
        default_value="Kusto.WebExplorer", 
        config=True, help="""Set the deep link destination.\n
        Abbreviation: 'qld'"""
    )

    plotly_fs_includejs = Bool(
        default_value=False,
        config=Dependencies.is_installed("plotly"),
        help="""Include plotly javascript code in popup window. If set to False (default), it download the script from https://cdn.plot.ly/plotly-latest.min.js.\n
        Abbreviation: 'pfi'"""
    )

    validate_connection_string = Bool(
        default_value=True, 
        config=True, 
        help="Validate connectionString with an implicit query, when query statement is missing. Abbreviation: vc"
    )

    auto_popup_schema = Bool(
        default_value=True, 
        config=True, 
        help="""Popup schema when connecting to a new database.\n
        Abbreviation: 'aps'"""
    )

    json_display = Enum(
        ["auto", "raw", "formatted"], 
        default_value="formatted", 
        config=True, 
        help="""Set json/dict display format.\n
        Abbreviation: 'jd'"""
    )

    schema_json_display = Enum(
        ["auto", "raw", "formatted"], 
        default_value="auto", 
        config=True, 
        help="""Set schema json/dict display format.\n
        Abbreviation: 'sjd'"""
    )

    palette_name = Unicode(
        default_value=Palettes.get_default_pallete_name(), 
        config=True, 
        help="""Set pallete by name to be used for charts.\n
        Abbreviation: 'pn'"""
    )

    palette_colors = Int(
        default_value=Palettes.DEFAULT_N_COLORS, 
        config=True, 
        help="""Set pallete number of colors to be used for charts.\n
        Abbreviation: 'pc'"""
    
    )

    palette_desaturation = Float(
        default_value=Palettes.DEFAULT_DESATURATION, 
        config=True, 
        help="""Set pallete desaturation to be used for charts.\n
        Abbreviation: 'pd'"""
    )

    temp_folder_name = Unicode(
        default_value=f"temp_files",
        read_only=True,
        config=True, 
        help=f"""Set the folder name for temporary files, relative to starting directory or user directory.\n
        Will be prefixed by {Constants.MAGIC_CLASS_NAME_LOWER}/ or .{Constants.MAGIC_CLASS_NAME_LOWER}/"""
    )

    temp_folder_location = Enum(
        ["auto", "starting_dir", "user_dir"], 
        default_value="auto",
        read_only=True,
        config=True, 
        help=f"""Set the location of the temp_folder, either within starting working directory or user workspace directory"""
    )

    # TODO: export files not used yet
    export_folder_name = Unicode(
        default_value=f"exported_files",
        read_only=True,
        config=True, 
        help=f"""Set the folder name for exported files, relative to starting directory or user directory.\n
        Will be prefixed by {Constants.MAGIC_CLASS_NAME_LOWER}/ or .{Constants.MAGIC_CLASS_NAME_LOWER}/"""
    )

    popup_interaction = Enum(
        ["auto", "button", "memory_button", "reference", "webbrowser_open_at_kernel", "reference_popup"],
        default_value="auto",
        config=True, 
        help="""Set popup interaction.\n
        Abbreviation: 'pi'"""        
    )

    temp_files_server = Enum(
        ["auto", "jupyter", "kqlmagic", "disabled"],
        default_value="auto" if Dependencies.is_installed("flask") else "disabled",
        read_only=True,
        config=Dependencies.is_installed("flask"),
        help="""Temp local files server."""        
    )

    temp_files_server_address = Unicode(
        default_value=None,
        read_only=True,
        config=Dependencies.is_installed("flask"), 
        allow_none=True, 
        help="""Temp files server address."""        
    )

    kernel_location = Enum(
        ["auto", "local", "remote"],
        default_value="auto",
        read_only=True,
        config=True, 
        help="""Kernel location"""    
    )

    cache_folder_name = Unicode(
        default_value=f"cache_files",
        read_only=True,
        config=True, 
        help=f"""Set the folder name for cache files, relative to starting directory or user directory.\n
        Will be prefixed by {Constants.MAGIC_CLASS_NAME_LOWER}/ or .{Constants.MAGIC_CLASS_NAME_LOWER}/"""
    )

    notebook_service_address = Unicode(
        default_value=None,
        read_only=True,
        config=True, 
        allow_none=True, 
        help="""Notebook service address."""        
    )

    notebook_app = Enum(
        ["auto", "jupyterlab", "azurenotebook", "azureml", "azuremljupyternotebook", "azuremljupyterlab", "jupyternotebook", "ipython", "visualstudiocode", "azuredatastudio", "azuredatastudiosaw", "nteract"], 
        default_value="auto",
        read_only=True,
        config=True, 
        help="""Set notebook application used."""
    )  # TODO: add "papermill"

    test_notebook_app = Enum(
        ["none", "jupyterlab", "azurenotebook", "azureml", "azuremljupyternotebook", "azuremljupyterlab", "jupyternotebook", "ipython", "visualstudiocode", "azuredatastudio", "azuredatastudiosaw", "nteract"], 
        default_value="none",
        read_only=True,
        config=True, 
        help="""Set testing application mode, results should return for the specified notebook application."""
    )  # TODO: add "papermill"

    kernel_id = Unicode(
        default_value=None,
        read_only=True,
        config=True,
        allow_none=True, 
        help="Current notebook kernel_id"
    )

    add_kql_ref_to_help = Bool(
        default_value=True,
        read_only=True,
        config=True, 
        help=f"""On {Constants.MAGIC_CLASS_NAME} load, auto add kql reference to Help menu."""
    )

    add_schema_to_help = Bool(
        default_value=True,
        read_only=True,
        config=True, 
        help="""On connection to database@cluster add  schema to Help menu."""
    )

    cache = Unicode(
        default_value=None, 
        read_only=True,
        config=True, 
        allow_none=True, 
        help="""Cache query results to be saved to the specified folder. (relative to cache_folder_name)"""
    )

    use_cache = Unicode(
        default_value=None,
        read_only=True,
        config=True, 
        allow_none=True, 
        help="""Use cached query results from the specified folder, instead of executing the query. (relative to cache_folder_name)"""
    )

    check_magic_version = Bool(
        default_value=True,
        read_only=True,
        config=True, 
        help=f"""On {Constants.MAGIC_CLASS_NAME} load, check whether new version of {Constants.MAGIC_CLASS_NAME} exist"""
    )

    show_what_new = Bool(
        default_value=True,
        read_only=True,
        config=True, 
        help=f"""On {Constants.MAGIC_CLASS_NAME} load, get history file of {Constants.MAGIC_CLASS_NAME} and show what new button to open it"""
    )

    show_init_banner = Bool(
        default_value=True,
        read_only=True,
        config=True, 
        help=f"""On {Constants.MAGIC_CLASS_NAME} load, show init banner"""
    )

    warn_missing_dependencies = Bool(
        default_value=True,
        read_only=True,
        config=True, 
        help=f"""On {Constants.MAGIC_CLASS_NAME} load, warn missing dependencies"""
    )

    kqlmagic_kernel = Bool(
        default_value=False,
        read_only=True,
        config=True, 
        help=f"""When set to True, {Constants.MAGIC_CLASS_NAME} kernel will be active"""
    )

    warn_missing_env_variables = Bool(
        default_value=True,
        read_only=True,
        config=True, 
        help=f"""On {Constants.MAGIC_CLASS_NAME} load, warn missing environment variables"""
    )

    debug = Bool(
        default_value=False,
        read_only=True,
        config=True, 
        help=f"""Used internally for debug only, when set to True, debug prints are displayed\n
        It is read on {Constants.MAGIC_CLASS_NAME} load, from environment variable {Constants.MAGIC_CLASS_NAME_UPPER}_DEBUG\n"""
    )

    allow_single_line_cell = Bool( 
        default_value=True,
        read_only=True,
        config=True,
        help=f"""When set to True, allows {Constants.MAGIC_CLASS_NAME} cell magic to include one line only, without body"""
    )

    allow_py_comments_before_cell = Bool( 
        default_value=True,
        read_only=True,
        config=True,
        help=f"""When set to True, allows {Constants.MAGIC_CLASS_NAME} cell magic to be prefixed by python comments"""
    )

    extras_require = Unicode(
        default_value=None,
        read_only=True,
        allow_none=True, 
        config=True, 
        help=f"""comma separated list of setup extras_require values that should be the same as specified at KqlmagicCustom install.\n
        It is read on {Constants.MAGIC_CLASS_NAME} load, from environment variable {Constants.MAGIC_CLASS_NAME_UPPER}_EXTRAS_REQUIRE\n
        """
    )

    request_id_tag = Unicode(
        default_value=None, 
        config=True, 
        allow_none=True, 
        help=f"""Tags request 'x-ms-client-request-id' header.\n
        Header pattern: {Constants.MAGIC_CLASS_NAME}.execute;{{tag}};{{guid}}\n
        Abbreviation: 'idtag'"""
    )

    request_app_tag = Unicode(
        default_value=None, 
        config=True, 
        allow_none=True, 
        help=f"""Tags request 'x-ms-app' header.\n
        Header pattern: {Constants.MAGIC_CLASS_NAME};{{tag}}\n
        Abbreviation: 'apptag'"""
    )

    request_user_tag = Unicode(
        default_value=None, 
        config=True, 
        allow_none=True, 
        help=f"""Tags request 'x-ms-user' header.\n
        Header pattern: {{tag}}\n
        Abbreviation: 'usertag'"""
    )
    
    request_user_agent_tag = Unicode(
        default_value=None, 
        config=True, 
        allow_none=True, 
        help=f"""Tags request 'User-Agent' header.\n
        Header pattern: {{tag}}\n
        Abbreviation: 'uatag'"""
    )

    request_cache_max_age = Int(
        default_value=0, 
        config=True, 
        allow_none=True, 
        help=f"""specifies, in seconds, the maximum amount of time a cached response is valid for.\n
        if set to 0, will bypass the response cache and always query the downstream services.\n
        if set to None, will use cached reponse as long it doesn't expires.\n
        Abbreviation: 'maxage'"""
    )

    # TODO: utilize using fig.show show(comfig=..) instead of offline.iplot
    plotly_config = _Dict(
        default_value=None,
        config=Dependencies.is_installed("plotly"),
        allow_none=True, 
        help=f"""plotly configuration options. see: https://plotly.com/python/configuration-options."""
    )

    dynamic_to_dataframe = Enum(
        ["object", "str"],
        default_value="object",
        config=Dependencies.is_installed("pandas"),
        help=f"""controls to what dataframe type should an kql dynamic value be translated.\n
        Abbreviation: 'dtd'"""
    )

    plotly_layout = _Dict(
        default_value=None, 
        config=Dependencies.is_installed("plotly"), 
        allow_none=True, 
        help=f"""plotly layout parameter, when set they override the defualt layout parameters.\n
        Abbreviation: 'pl'"""        
    )

    auth_token_warnings = Bool(
        default_value=False, 
        config=True, 
        help=f"""When set to True, will display auth token warning when token different from connection string params.\n
        Abbreviation: 'atw'"""
    )

    auth_use_http_client = Bool(
        default_value=False,
        read_only=True,
        config=True, 
        help=f"""When set to True, {Constants.MAGIC_CLASS_NAME} http_client will be used by msal"""
    )

    enable_curly_brackets_params = Bool(
        default_value=False, 
        config=True,
        help=f"""when set to True, strings within curly brackets will be evaluated as a python expression, and if evaluation succeeds result will replace the string (including the curly brackets).\n
        If evaluation fails it will stay as is, including the curly brackets.\n
        To escape Curly brackets the must be doubled {{{{something}}}}\n
        Abbreviation: 'ecbp'"""
    )

    logger().debug("Kqlmagic:: - define class code")


    @staticmethod
    def get_default_options_dict()->Dict[str,Any]:
        c = kql_core_obj.default_options
        return {name: getattr(c, name) for name in c.class_traits() if c.trait_metadata(name, "config")}


    @staticmethod
    def get_default_options()->Configurable:
        return kql_core_obj.default_options


    @validate("auto_dataframe")
    def _valid_value_auto_dataframe(self, proposal:Dict[str,Any]):
        try:
            if proposal["value"] is True and not Dependencies.is_installed("pandas"):
                raise ValueError("cannot be set to True, 'pandas' package is not installed")
        except (AttributeError, ValueError) as e:
            message = f"The 'auto_dataframe' trait of a {Constants.MAGIC_CLASS_NAME} instance {str(e)}"
            raise TraitError(message)
        return proposal["value"]


    @validate("palette_name")
    def _valid_value_palette_name(self, proposal:Dict[str,Any]):
        try:
            Palette.validate_palette_name(proposal["value"])
        except (AttributeError, ValueError) as e:
            message = f"The 'palette_name' trait of a {Constants.MAGIC_CLASS_NAME} instance {str(e)}"
            raise TraitError(message)
        return proposal["value"]


    @validate("palette_desaturation")
    def _valid_value_palette_desaturation(self, proposal:Dict[str,Any]):
        try:
            Palette.validate_palette_desaturation(proposal["value"])
        except (AttributeError, ValueError) as e:
            message = f"The 'palette_desaturation' trait of a {Constants.MAGIC_CLASS_NAME} instance {str(e)}"
            raise TraitError(message)
        return proposal["value"]


    @validate("palette_colors")
    def _valid_value_palette_color(self, proposal:Dict[str,Any]):
        try:
            Palette.validate_palette_colors(proposal["value"])
        except (AttributeError, ValueError) as e:
            message = f"The 'palette_color' trait of a {Constants.MAGIC_CLASS_NAME} instance {str(e)}"
            raise TraitError(message)
        return proposal["value"]


    @validate("notebook_app")
    def _valid_value_notebook_app(self, proposal:Dict[str,Any]):
        try:
            if proposal["value"] == "auto":
                raise ValueError("cannot be set to auto, after instance is loaded")
        except (AttributeError, ValueError) as e:
            message = f"The 'notebook_app' trait of a {Constants.MAGIC_CLASS_NAME} instance {str(e)}"
            raise TraitError(message)
        return proposal["value"]


    @validate("table_package")
    def _valid_value_table_package(self, proposal:Dict[str,Any]):
        try:
            if proposal["value"] == "auto":
                raise ValueError("cannot be set to auto, after instance is loaded")
        except (AttributeError, ValueError) as e:
            message = f"The 'table_package' trait of a {Constants.MAGIC_CLASS_NAME} instance {str(e)}"
            raise TraitError(message)
        return proposal["value"]     


    @validate("temp_folder_location")
    def _valid_value_temp_folder_location(self, proposal:Dict[str,Any]):
        try:
            if proposal["value"] == "auto":
                raise ValueError("cannot be set to auto, after instance is loaded")
        except (AttributeError, ValueError) as e:
            message = f"The 'temp_folder_location' trait of a {Constants.MAGIC_CLASS_NAME} instance {str(e)}"
            raise TraitError(message)
        return proposal["value"]   


    @validate("temp_files_server")
    def _valid_value_temp_files_server(self, proposal:Dict[str,Any]):
        try:
            if proposal["value"] != self.temp_files_server:
                if self.temp_files_server == "disabled":
                    raise ValueError("feature is 'disabled', due to missing 'flask' module")
        except (AttributeError, ValueError) as e:
            message = f"The 'temp_files_server' trait of a {Constants.MAGIC_CLASS_NAME} instance {str(e)}"
            raise TraitError(message)
        return proposal["value"]


    @validate("kernel_id")
    def _valid_value_kernel_id_app(self, proposal:Dict[str,Any]):
        try:
            if self.kernel_id is not None:
                raise ValueError("cannot be set, it is readonly, set internally")
        except (AttributeError, ValueError) as e:
            message = f"The 'kernel_id' trait of a {Constants.MAGIC_CLASS_NAME} instance {str(e)}"
            raise TraitError(message)
        return proposal["value"]


    @validate("try_msi")
    def _valid_value_try_msi(self, proposal:Dict[str,Any]):
        try:
            msi_params = proposal["value"]
            if msi_params is not None:
                valid_params = ["port", "timeout", "client_id", "object_id", "msi_res_id", "cloud_environment", "resource"]
                for key in msi_params:
                    if key not in valid_params:
                        raise ValueError(f"unknown param '{key}'. Supported params: {valid_params}")
                exclusive_pcount = 0
                for key in ["client_id", "object_id", "msi_res_id"]:
                    if msi_params.get(key) is not None:
                        exclusive_pcount += 1
                if exclusive_pcount > 1:
                    raise ValueError("the following parameters are mutually exclusive and can not be provided at the same time: user_uid, object_id, msi_res_id")
        except (AttributeError, ValueError) as e:
            message = f"The 'try_msi' trait of a {Constants.MAGIC_CLASS_NAME} instance {str(e)}"
            raise TraitError(message)
        return proposal["value"]


    @validate("try_token")
    def _valid_value_try_token(self, proposal:Dict[str,Any]):
        try:
            token = proposal["value"]
            if token is not None:
                mandatory_properties = [["tokenType", "token_type"], ["accessToken", "access_token"]]
                for pair in mandatory_properties:
                    if token.get(pair[0]) is None and token.get(pair[1]) is None:
                        raise ValueError(f"one of '{pair}' property is mandatory and is not set in token. mandatory properties are: {mandatory_properties}")                
        except (AttributeError, ValueError) as e:
            message = f"The 'try_token' trait of a {Constants.MAGIC_CLASS_NAME} instance {str(e)}"
            raise TraitError(message)
        return proposal["value"]


    def __init__(self, shell, global_ns=None, local_ns=None, is_magic=True)->None:
        global kql_core_obj, kql_core_count
        kql_core_count += 1
        if kql_core_obj is None:
            Configurable.__init__(self, config=(shell.config if shell is not None else None))
            
            
            self._config_traits = self.traits()
            self._read_only_config_trait_names = self._get_read_only_config_trait_names()
            self._set_read_only_config_traits(False)
            # Add ourself to the list of module configurable via %config
            if shell is not None:
                shell.configurables.append(self)
            default_options = self
        else:
            default_options = kql_core_obj.default_options

        default_options.set_trait("is_kernel_intializtion", Kqlmagic.is_kqlmagic_kernel, force=True)

        OsDependentAPI(default_options)

        if is_magic and is_magics_class and self.is_ipython_extension:
            Magics.__init__(self, shell=shell)
            default_options.set_trait("is_magic", True, force=True)
        else:
            default_options.set_trait("is_magic", False, force=True)
            default_options.set_trait("show_init_banner", False, force=True, lock=True)

        if kql_core_obj is None:
            kql_core_obj = Kqlmagic_core(global_ns=global_ns, local_ns=local_ns, shell=shell, default_options=self)
            self._set_read_only_config_traits(True)

        default_options.set_trait("is_kernel_intializtion", False, force=True)


    @property
    def config_traits(self)->Dict[str,TraitType]:
        return self._config_traits


    @property
    def read_only_trait_names(self)->List[str]:
        return self._read_only_config_trait_names


    def _get_read_only_config_trait_names(self):
        return [name for name, trait in self._config_traits.items() if trait.read_only==True]


    def _set_read_only_config_traits(self, is_read_only):
        for name in self._read_only_config_trait_names:
            self._config_traits[name].read_only = is_read_only == True


    def set_trait(self, name:str, value:Any, force:bool=False, lock:bool=False):
        if force is True:
            super(Kqlmagic, self).set_trait(name, value)
        else:
            try:
                setattr(self, name, value)
            except TraitError:
                if value != getattr(self, name):
                    raise
            
        if lock is True:
            self._config_traits[name].read_only = True
        return getattr(self, name)


    def is_read_only(self, name:str)->bool:
        return self._config_traits[name].read_only


    @needs_local_scope
    @line_magic(Constants.MAGIC_NAME)
    @cell_magic(Constants.MAGIC_NAME)
    def execute(self, 
                line:str, 
                cell:str=None, 
                local_ns:dict={}, 
                override_vars:Dict[str,str]=None, 
                override_options:Dict[str,Any]=None, 
                override_query_properties:Dict[str,Any]=None, 
                override_connection:str=None, 
                override_result_set:ResultSet=None):

        # Known issue:
        #
        # first line of magic is auto expanded by shell, before it reaches magic code !!!
        # ask users to avoid using curly brackets in line magics, and in fisrt line of cell magic
        # if required put double brackets
        #

        result = kql_core_obj.execute(
            line=line, 
            cell=cell, 
            local_ns=local_ns,
            override_vars=override_vars,
            override_options=override_options,
            override_query_properties=override_query_properties,
            override_connection=override_connection,
            override_result_set=override_result_set)

        return result

    @staticmethod
    def stop(unload_ipython_extension=False, kql_stop=False)->None:
        global kql_core_obj, is_non_magic_kql_on, kql_core_count
        if kql_core_obj and kql_core_count > 0:
            if unload_ipython_extension is True:
                default_options = kql_core_obj.default_options
                default_options.set_trait("is_magic", False, force=True)

            kql_core_count -= 1
            if kql_core_count == 0:
                kql_core_obj.stop()
                kql_core_obj = None
                is_non_magic_kql_on = False
            elif kql_stop is True:
                is_non_magic_kql_on = False
        

def kql(text:str='', options:Dict[str,Any]=None, query_properties:Dict[str,Any]=None, vars:Dict[str,str]=None, connection_string:str=None, global_ns=None, local_ns=None):
    global kql_core_obj, is_non_magic_kql_on, kql_core_count

    if not is_non_magic_kql_on:
        if kql_core_obj is None:
            shell = None
            if global_ns is None and local_ns is None:
                shell = IPythonAPI.get_shell()
                if shell is None:
                    global_ns = globals()
                    local_ns = locals()
            
            Kqlmagic(
                shell=shell, 
                global_ns=global_ns, 
                local_ns=local_ns, 
                is_magic=False)
        else:
            kql_core_count += 1
        is_non_magic_kql_on = True

    if text.find("\n"):
        line = ""
        cell = text
    else:
        line = text
        cell = None

    return kql_core_obj.execute(
        line, 
        cell,
        override_vars=vars,
        override_options=options, 
        override_query_properties=query_properties, 
        override_connection=connection_string)


def kql_stop():
    if is_non_magic_kql_on:
        Kqlmagic.stop(kql_stop=True)


# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

from datetime import datetime
import json
import string
import random
import threading


from .dependencies import Dependencies
from .my_utils import json_dumps
from .sso_storage import get_sso_store
from .dict_db_storage import DictDbStorage
from .log import logger


msal = Dependencies.get_module("msal", dont_throw=True)


if msal:
    class MsalTokenCache(msal.SerializableTokenCache):

        @classmethod
        def get_cache(cls, cache_selector_key: str, **options)-> dict:
            "retreive the cache from store, based the selector key"
            store = get_sso_store(cache_selector_key, **options)
            if store:
                cache = MsalTokenCache(store)
                return cache


        def __init__(self, store:DictDbStorage, state:str=None):
            super(MsalTokenCache, self).__init__()
            self._rlock = threading.RLock()

            self._cache_state = super(MsalTokenCache, self).serialize()
            self._store = store
            self.deserialize(state=state)


        # def find(self, credential_type: str, target: list=None, query: dict=None):
        def find(self, *args, **kwargs):
            '''find entries in cache'''
            logger().debug(f"MsalTokenCache find({args}, {kwargs})")

            with self._rlock:
                self.deserialize()
                result = super(MsalTokenCache, self).find(*args, **kwargs)
                return result


        # def modify(self, credential_type: str, old_entry:dict, new_key_value_pairs: dict=None):
        def modify(self, *args, **kwargs):
            '''modify or remove entries in cache'''
            logger().debug(f"MsalTokenCache modify({args}, {kwargs})")

            with self._rlock:
                self.deserialize()
                super(MsalTokenCache, self).modify(*args, **kwargs)
                self.serialize()


        def _random_string(self) -> str:
            return ''.join(random.choice(string.ascii_uppercase + string.ascii_lowercase + string.digits) for _ in range(random.randint(1, 100)))


        def serialize(self) -> str:
            '''serialize cache'''

            with self._rlock:
                new_cache_state = super(MsalTokenCache, self).serialize()
                if new_cache_state != self._cache_state:
                    self._cache_state = new_cache_state
                    state_obj = {
                        "description": "kqlmagic",
                        "version": 2,
                        "package": "msal",
                        "timestamp": int(datetime.utcnow().timestamp()),
                        "random_string": self._random_string(),  # makes length and content different each time
                        "cache_state": self._cache_state,
                    }
                    state = json_dumps(state_obj)
                    self._store.save(state)
                    return state


        def deserialize(self, state: str=None):
            '''deserialize cache'''

            with self._rlock:
                state = state or self._store.restore()
                if state:
                    state_obj = json.loads(state)
                    if (state_obj.get("description") == "kqlmagic"
                            and state_obj.get("version") == 2
                            and state_obj.get("package") == "msal"
                            and state_obj.get("timestamp") < int(datetime.utcnow().timestamp())
                            and state_obj.get("random_string")
                            and len(state_obj.get("random_string")) >= 1 
                            and len(state_obj.get("random_string")) <= 100 
                            and state_obj.get("cache_state") is not None):
                        self._cache_state = state_obj.get("cache_state")
                        super(MsalTokenCache, self).deserialize(self._cache_state)

else:
    class MsalTokenCache(object):

        @classmethod
        def get_cache(cls, cache_selector_key: str, **options)-> dict:
            return None



#!/usr/bin/env python

#-------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
#--------------------------------------------------------------------------
"""Setup for Kqlmagic"""

import sys
import os
import io


# Always prefer setuptools over distutils
from setuptools import setup, find_packages



#==============================================================================
# _IS_CUSTOM_SETUP VALUE IS MODIFIED BY A SHELL PREPROCESSOR SCRIPT
# TO FORK TWO PACKAGES:
#      - Kqlmagic - INCLUDES ALL DEPENDENCIES
#      - KqlmagicCustom - DEPENDENCIES NEED TO BE EXPLICITLY SPECIFIED
#==============================================================================
# NEXT CODE LINE SHOULD NOT BE TOUCHED/MODIFIED MANUALLY, MAKE SURE IT APPEARS ONLY ONCE
_IS_CUSTOM_SETUP = False # MODIFIED BY A SHELL PREPROCESSOR SCRIPT
#==============================================================================
#
#

_PACKAGE_DASHED_PATH   = 'azure-Kqlmagic'

_VERSION_CODE_FILENAME = '_version.py'
_REQUIRE_CODE_FILENAME = '_require.py'
_META_CODE_FILENAME    = '_meta.py'
_TEST_CODE_FILENAME    = '_test.py'


#
# help functions
#

# Union Without repetition  
def list_union(*args): 
    final_list = list(set().union(*args)) 
    return final_list 


def read(relative_path, encoding='utf-8'):
    path = os.path.join(os.path.dirname(__file__), relative_path)
    with io.open(path, encoding=encoding) as fp:
        return fp.read()


def write(content, relative_path, encoding='utf-8'):
    path = os.path.join(os.path.dirname(__file__), relative_path)
    with io.open(path, mode="w", encoding=encoding) as fp:
        return fp.write(content)


def exec_py_code(filename):
    """Get requires from _require.py file."""

    code_relative_path = _PACKAGE_DASHED_PATH.replace('-', '/') + '/' + filename
    try:
        package_relative_path = _PACKAGE_DASHED_PATH.replace('-', os.path.sep)

        code_relative_path = os.path.join(package_relative_path, filename)
        py_code = read(code_relative_path)
        user_ns = {}
        exec(py_code, {}, user_ns)
        return user_ns
    except Exception as error:
        raise RuntimeError("Failed to execute py code from file: {} due to error: {}".format(code_relative_path, error))


#
# get version
#

def get_version():
    """Get version and devstatus from _version.py file."""
    user_ns = exec_py_code(_VERSION_CODE_FILENAME)
    # Get development Status for classifiers
    version = user_ns["__version__"]
    version_info = user_ns["__version_info__"]
    devstatus = {
        'dev':   '2 - Pre-Alpha',
        'alpha': '3 - Alpha',
        'beta':  '4 - Beta',
        'rc':    '4 - Beta',
        'final': '5 - Production/Stable',
        'post':  '5 - Production/Stable'
    }[version_info[3]]
    return version, devstatus

VERSION, _DEVSTATUS = get_version()


#
# get package metadata
#

def get_meta(is_custom_setup, devstatus):
    """Get metadata from _meta.py file."""
    meta = exec_py_code(_META_CODE_FILENAME)
    if is_custom_setup:
        meta["NAME"] = meta["NAME"] + "Custom"
    meta["CLASSIFIERS"].insert(0, 'Development Status :: {}'.format(devstatus))
    return meta

meta = get_meta(_IS_CUSTOM_SETUP, _DEVSTATUS)

DESCRIPTION         = meta["DESCRIPTION"]
NAME                = meta['NAME']
AUTHOR              = meta["AUTHOR"]
AUTHOR_EMAIL        = meta["AUTHOR_EMAIL"]
MAINTAINER          = meta["MAINTAINER"]
MAINTAINER_EMAIL    = meta["MAINTAINER_EMAIL"]
URL                 = meta["URL"]
PROJECT_URLS        = meta["PROJECT_URLS"]
LICENSE             = meta["LICENSE"]
KEYWORDS            = meta["KEYWORDS"]
CLASSIFIERS         = meta["CLASSIFIERS"]
PYTHON_REQUIRES     = meta["PYTHON_REQUIRES"]


#
# check minimal sanity
#

#==============================================================================
# Minimal Python version sanity check
# Taken from the notebook setup.py -- Modified BSD License
#==============================================================================
def minimal_sanity(python_requires, kqlmagic_name, kqlmagic_version):

    def get_version_tuple(item:str):
        "parse package name from a package line that include version info"
        ver_tuple = (3, 6)
        item = item.strip().lower()
        for i in range(len(item)):
            if item[i].isdigit():
                ver_tuple = [int(val) for val in item[i:].split(".")]
                break
        return ver_tuple

    v = sys.version_info
    min_python_required_ver = get_version_tuple(python_requires)
    if list(v[:2]) < min_python_required_ver:
        pip_message = 'This may be due to an out of date pip. Make sure you have pip >= 9.0.1.'
        try:
            import pip
            pip_version = tuple([int(x) for x in pip.__version__.split('.')[:3]])
            if pip_version < (9, 0, 1) :
                pip_message = 'Your pip version is out of date, please install pip >= 9.0.1. '\
                'pip {} detected.'.format(pip.__version__)
            else:
                # pip is new enough - it must be something else
                pip_message = ''
        except Exception:
            pass

        error = """ERROR: {name} {ver} requires Python version 3.6 or above.

        Python {py} detected.
        {pip}
        """.format(name=kqlmagic_name, ver=kqlmagic_version, py=sys.version_info, pip=pip_message)
        print(error, file=sys.stderr)
        sys.exit(1)

minimal_sanity(PYTHON_REQUIRES, NAME, VERSION)


#
# get long description
#

def get_long_description():
    """Get long description from REAM.RST and NEWS.txt."""
    try:
        readme_rst_content = read('README.rst')
        long_description_content_type = 'text/x-rst' # long_description_content_type = 'text/markdown'
        news_txt_content = read('NEWS.txt')
        long_description =  (readme_rst_content + '\n\n' + news_txt_content).replace('\r','')
        return long_description, long_description_content_type
    except Exception as error:
        raise RuntimeError("Unable to README.rst or NEWS.txt file, due to error: {}".format(error))

LONG_DESCRIPTION, LONG_DESCRIPTION_CONTENT_TYPE = get_long_description()


#
# get required dependencies
#

def get_require():
    user_ns = exec_py_code(_REQUIRE_CODE_FILENAME)
    return user_ns["INSTALL_REQUIRES"], user_ns["EXTRAS_REQUIRE"], user_ns["DEV_REQUIRES"], user_ns["TESTS_REQUIRE"] # pylint: disable=no-member

INSTALL_REQUIRES, EXTRAS_REQUIRE, DEV_REQUIRES, TESTS_REQUIRE = get_require()

if _IS_CUSTOM_SETUP:
    DESCRIPTION = DESCRIPTION + " (Custom Dependencies)"
else:
    INSTALL_REQUIRES = list_union(INSTALL_REQUIRES, EXTRAS_REQUIRE.get("default"))
    EXTRAS_REQUIRE   = {}

EXTRAS_REQUIRE["dev"] = DEV_REQUIRES
EXTRAS_REQUIRE["tests"] = TESTS_REQUIRE


#
# get test command class
#

def get_test_class():
    user_ns = exec_py_code(_TEST_CODE_FILENAME)
    return user_ns["TestClass"] # pylint: disable=no-member

TEST_CLASS = get_test_class()


def kwargs_params_to_dict(**kwargs):
    import json
    print(json.dumps(kwargs, default=lambda o: str(o), indent=4))
    return kwargs

setup_kwargs = kwargs_params_to_dict(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type=LONG_DESCRIPTION_CONTENT_TYPE,
    classifiers=CLASSIFIERS,
    keywords=KEYWORDS,
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    maintainer=MAINTAINER,
    maintainer_email=MAINTAINER_EMAIL,
    url=URL,
    license=LICENSE,
    python_requires=PYTHON_REQUIRES,
    packages=find_packages('azure'),
    package_dir={'': 'azure'},
    include_package_data=True,
    zip_safe=False,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
    project_urls=PROJECT_URLS,
    cmdclass={'test': TEST_CLASS},
    tests_require=TESTS_REQUIRE,
)

setup(**setup_kwargs)



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import logging
import datetime
import uuid
import traceback
from typing import Any, List, Dict


from .constants import Constants
from .my_utils import get_env_var
from .ipython_api import IPythonAPI


isNullLogger:bool = False


def initialize(log_level:str=None, log_file:str=None, log_file_prefix:str=None, log_file_mode:str=None)->None:
    global isNullLogger
    log_level = log_level or get_env_var(f"{Constants.MAGIC_CLASS_NAME_UPPER}_LOG_LEVEL")
    log_file = log_file or get_env_var(f"{Constants.MAGIC_CLASS_NAME_UPPER}_LOG_FILE")
    log_file_prefix = log_file_prefix or get_env_var(f"{Constants.MAGIC_CLASS_NAME_UPPER}_LOG_FILE_PREFIX")
    log_file_mode = log_file_mode or get_env_var(f"{Constants.MAGIC_CLASS_NAME_UPPER}_LOG_FILE_MODE")
    if log_level or log_file or log_file_mode or log_file_prefix:
        kernel_id = IPythonAPI.get_notebook_kernel_id() or "kernel_id"

        log_level = log_level or logging.DEBUG
        log_file = log_file or f"{log_file_prefix or 'kqlmagic'}-{kernel_id}.log"
        # handler's default mode is 'a' (append)
        log_file_mode = (log_file_mode or "w").lower()[:1]
        log_handler = logging.FileHandler(log_file, mode=log_file_mode)
        isNullLogger = False
    else:
        log_handler = logging.NullHandler()
        isNullLogger = True

    set_logging_options({'level': log_level, 'handler': log_handler})
    set_logger(Logger())

    if log_file:
        if log_file_mode == "a":
            logger().debug("\n\n----------------------------------------------------------------------")
        now = datetime.datetime.now()

        logger().debug(f"start date {now.isoformat()}\n")
        logger().debug(f"logger level {log_level}\n")
        logger().debug("logger init done")


def create_log_context(correlation_id:str=None)->Dict[str,str]:
    return {"correlation_id": correlation_id or str(uuid.uuid4())}


def set_logging_options(options:Dict[str,Any]=None)->None:
    """Configure logger, including level and handler spec'd by python
    logging module.

    Basic Usages::
        >>>set_logging_options({
        >>>  'level': 'DEBUG'
        >>>  'handler': logging.FileHandler(<file-name>) # file name can be 
        >>>})
    """
    if options is None:
        options = {}
    logger = logging.getLogger(Constants.LOGGER_NAME)

    logger.setLevel(options.get("level", logging.ERROR) or logging.ERROR)

    handler = options.get("handler")
    if handler:
        handler.setLevel(logger.level)
        logger.addHandler(handler)


def get_logging_options()->Dict[str,Any]:
    """Get logging options

    :returns: a dict, with a key of 'level' for logging level.
    """
    logger = logging.getLogger(Constants.LOGGER_NAME)
    level = logger.getEffectiveLevel()
    return {"level": logging.getLevelName(level)}


#    log_context = log.create_log_context(correlation_id)
#    logger = log.Logger('SomeComponent', log_context)
class Logger(object):
    """wrapper around python built-in logging to log correlation_id, and stack
    trace through keyword argument of 'log_stack_trace'
    """

    def __init__(self, component_name:str=None, log_context:Dict[str,str]=None)->None:
        # if not log_context:
        #     raise AttributeError('Logger: log_context is a required parameter')

        self._component_name = component_name
        self.log_context = log_context
        self._logging = logging.getLogger(Constants.LOGGER_NAME)
        self._current_log_buffer = []


    def _log_message(self, msg:str, log_stack_trace:bool=None)->str:
        formatted = ""

        if self.log_context:
            correlation_id = self.log_context.get("correlation_id")
            if correlation_id:
                formatted = f"{correlation_id} - "

        if self._component_name:
            formatted += f"{self._component_name}:"

        formatted += msg

        if log_stack_trace:
            formatted += f"\nStack:\n{traceback.format_stack()}"

        self._current_log_buffer.append(formatted)
        return formatted
        

    def isNull(self)->bool:
        return isNullLogger


    def getCurrentLogMessages(self)->List[str]:
        return self._current_log_buffer


    def resetCurrentLogMessages(self)->None:
        self._current_log_buffer = []


    def critical(self, msg:str, *args, **kwargs)->None:
            try:
                log_stack_trace = kwargs.pop("log_stack_trace", None)
                msg = self._log_message(msg, log_stack_trace)
                self._logging.critical(msg, *args, **kwargs)
            except Exception as e:
                self._logException(e, "critical")


    def error(self, msg:str, *args, **kwargs)->None:
        try:
            log_stack_trace = kwargs.pop("log_stack_trace", None)
            msg = self._log_message(msg, log_stack_trace)
            self._logging.error(msg, *args, **kwargs)
        except Exception as e:
            self._logException(e, "error")


    def warn(self, msg:str, *args, **kwargs)->None:
        try:
            log_stack_trace = kwargs.pop("log_stack_trace", None)
            msg = self._log_message(msg, log_stack_trace)
            self._logging.warning(msg, *args, **kwargs)
        except Exception as e:
            self._logException(e, "warn")


    def info(self, msg:str, *args, **kwargs)->None:
        try:
            log_stack_trace = kwargs.pop("log_stack_trace", None)
            msg = self._log_message(msg, log_stack_trace)
            self._logging.info(msg, *args, **kwargs)
        except Exception as e:
            self._logException(e, "info")


    def debug(self, msg:str, *args, **kwargs)->None:
        try:
            log_stack_trace = kwargs.pop("log_stack_trace", None)
            msg = self._log_message(msg, log_stack_trace)
            self._logging.debug(msg, *args, **kwargs)
        except Exception as e:
            self._logException(e, "debug")


    def exception(self, msg:str, *args, **kwargs)->None:
        try:
            log_stack_trace = kwargs.pop("log_stack_trace", None)
            msg = self._log_message(msg, log_stack_trace)
            self._logging.exception(msg, *args, **kwargs)
        except Exception as e:
            self._logException(e, "exception")


    def _logException(self, ex:Exception, level:str, **kwargs)->None:
        try:
            log_stack_trace = kwargs.pop("log_stack_trace", None)
            msg = f"failed to log level {level} record"
            try:
                msg = self._log_message(msg, log_stack_trace)
            except:
                pass
            self._logging.exception(msg, exc_info=ex)
        except:
            pass


current_logger:Logger = None


def logger()->Logger:
    return current_logger


def set_logger(new_logger:Logger)->Logger:
    global current_logger
    current_logger = new_logger
    return current_logger


initialize()



# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

import os
import sys
import platform
import subprocess

from typing import Any, Callable, Iterable


from ._debug_utils import debug_print
from traitlets.config.configurable import Configurable


class OsDependentAPI(object):
    """
    """

    options:Configurable = None

    def __init__(self, options:Configurable )->None:
        self.options = options
    
    
    @classmethod
    def webbrowser_open(cls, url:str)->None:
        # return webbrowser.open(url)
        # webbrowser.open(url, new=1, autoraise=True)
        if cls.options.get("notebook_app") in ["azuredatastudiosaw"]:
            try:
                OsDependentAPI.startfile(url)
                return
            except:
                pass
        try:
            import webbrowser
            webbrowser.open(url)
        except:
            OsDependentAPI.startfile(url)


    @classmethod
    def startfile(filename:str)->None:
        platform = sys.platform
        if platform[:3] == "win":
            os.startfile(filename)
        else:
            opener = "open" if platform == "darwin" else "xdg-open"
            subprocess.call([opener, filename])


# -------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for
# license information.
# --------------------------------------------------------------------------

# DPAPI access library
# This file uses code originally created by Crusher Joe:
# http://article.gmane.org/gmane.comp.python.ctypes/420
#

import platform
import getpass


if platform.system() == 'Windows':
    try:
        from msal_extensions.windows import WindowsDataProtectionAgent
    except Exception:
        dpapi_installed = False
    else:
        dpapi_installed = True
else:
    dpapi_installed = False


from .log import logger
from .constants import DpapiParam, SsoCrypto    


if dpapi_installed:

    class DpapiCrypto(object):

        def __init__(self, **crypto_options)->None:
            entropy: str = crypto_options.get(DpapiParam.ENTROPY) or "kqlmagic"
            self._description: str = crypto_options.get(DpapiParam.DESCRIPTION)
            self._dp_agent = WindowsDataProtectionAgent(entropy=entropy)


        @property
        def suffix(self)->str:
            return getpass.getuser()


        def encrypt(self, data:str)->bytes:
            logger().debug(f"DpapiCrypto(object)::encrypt(self, data)")
            if data:
                return self._dp_agent.protect(data)


        def decrypt(self, encrypted_data_bytes:bytes)->str:
            logger().debug(f"DpapiCrypto(object)::decrypt(self, encrypted_data_bytes)")
            if encrypted_data_bytes:
                return self._dp_agent.unprotect(encrypted_data_bytes)


        def verify(self, encrypted_data_bytes:bytes)->None:
            pass

else:
    class DpapiCrypto(object):

        def __init__(self, **crypto_options):
            raise Exception(f"Warning: SSO is not activated due to {SsoCrypto.DPAPI} cryptography failed to import WindowsDataProtectionAgent from msal_extensions.windows")
